import functools
from typing import NamedTuple, Optional, Union

import jax
import jax.numpy as jnp
import numpy as np
from jax import lax
from jax.experimental import pallas as pl
from jax.experimental.pallas import tpu as pltpu

DEFAULT_MASK_VALUE = -0.7 * float(np.finfo(np.dtype("float32")).max)

P = jax.sharding.PartitionSpec
MAX_INT8 = 127.5


class QuantizedTensor(NamedTuple):
	"""A tensor which has been quantized to int8 and its scales.

	Attributes:
	  weight: Weight
	  scales: Scales
	"""

	weight: jnp.ndarray
	scales: jnp.ndarray


def to_int8(x: jnp.ndarray, h: jnp.ndarray) -> jnp.ndarray:
	"""Converts a float array to an int8 array with a scale.

	Args:
	  x: Float array.
	  h: Quantization scale.

	Returns:
	  Int8 array.
	"""
	return jnp.int8(jnp.rint(x * (MAX_INT8 / h)))


def from_int8(
	x: jnp.ndarray, h: jnp.ndarray, dtype: jnp.dtype = jnp.bfloat16
) -> jnp.ndarray:
	"""Converts an int8 array to a float array with a scale.

	Args:
	  x: Int8 array.
	  h: Quantization scale.
	  dtype: Float dtype to convert to.

	Returns:
	  Float array.
	"""
	return x.astype(dtype) * h / MAX_INT8


def get_quantization_scales(x: jnp.ndarray) -> jnp.ndarray:
	"""Computes the quantization scales for a float array.

	These are the maximum values of the trailing dimension.

	Args:
	  x: Float array to quantize.

	Returns:
	  Array of the same shape as input but with the trailing dimension reduced to
	  a size 1 absolute max value.
	"""
	return jnp.max(jnp.abs(x), axis=-1, keepdims=True)


def quantize_to_int8(
	x: jnp.ndarray,
) -> QuantizedTensor:
	"""Quantizes a float array to an int8 QuantizedTensor.

	Args:
	  x: Float array to quantize.

	Returns:
	  Int8 QuantizedTensor.
	"""
	x_scales = get_quantization_scales(x)
	return QuantizedTensor(weight=to_int8(x, x_scales), scales=x_scales)


def unquantize_from_int8(
	x: QuantizedTensor,
	dtype: jnp.dtype = jnp.bfloat16,
) -> jnp.ndarray:
	"""Unquantizes an int8 QuantizedTensor to a float array.

	Args:
	  x: Int8 QuantizedTensor to unquantize.
	  dtype: Float dtype to unquantize to.

	Returns:
	  Float array.
	"""
	return from_int8(x.weight, x.scales, dtype)


class MultiPageAsyncCopyDescriptor:
	"""Descriptor for async copy of multiple K/V pages from HBM."""

	def __init__(
		self,
		pages_hbm_ref,
		scales_pages_hbm_ref,
		vmem_buffer,
		scales_vmem_buffer,
		sem,
		page_indices,
		page_indices_start_offset,
		num_pages_to_load,
		head_index,
	):
		self._vmem_buffer = vmem_buffer
		self._scales_vmem_buffer = scales_vmem_buffer
		self._num_pages_to_load = num_pages_to_load
		if head_index is not None:
			self._pages_hbm_ref = pages_hbm_ref.at[head_index]
			if scales_pages_hbm_ref is not None:
				self._scales_pages_hbm_ref = scales_pages_hbm_ref.at[head_index]
			else:
				self._scales_pages_hbm_ref = None
		else:
			self._pages_hbm_ref = pages_hbm_ref
			self._scales_pages_hbm_ref = scales_pages_hbm_ref
		self._sem = sem
		self._page_indices = page_indices
		self._page_indices_start_offset = page_indices_start_offset
		self._async_copies = [
			self._make_async_copy(i) for i in range(self._num_pages_to_load)
		]
		if self._scales_pages_hbm_ref is not None and self._scales_vmem_buffer is not None:
			self._async_copies += [
				self._make_scales_async_copy(i) for i in range(self._num_pages_to_load)
			]

	def _make_async_copy(self, i):
		page_index = self._page_indices[self._page_indices_start_offset + i]
		return pltpu.make_async_copy(
			self._pages_hbm_ref.at[page_index], self._vmem_buffer.at[i], self._sem
		)

	def _make_scales_async_copy(self, i):
		page_index = self._page_indices[self._page_indices_start_offset + i]
		return pltpu.make_async_copy(
			self._scales_pages_hbm_ref.at[page_index],  # pytype: disable=attribute-error
			self._scales_vmem_buffer.at[i],  # pytype: disable=attribute-error
			self._sem,
		)

	def start(self):
		"""Starts the async copies."""
		for async_copy in self._async_copies:
			async_copy.start()

	def _maybe_dequantize(self, x, x_scale, dtype=jnp.bfloat16):
		if x_scale is None:
			return x.astype(dtype)
		return from_int8(x, x_scale, dtype=dtype)

	def wait_and_get_loaded(self) -> jax.Array:
		"""Wait async copies and gets the loaded buffer as a jax.Array."""
		for async_copy in self._async_copies:
			async_copy.wait()
		head_dim = self._vmem_buffer.shape[-1]
		jax_array = self._vmem_buffer[...].astype(jnp.float32)
		if self._scales_vmem_buffer is not None:
			scales_jax_array = self._scales_vmem_buffer[...].astype(jnp.float32)
		else:
			scales_jax_array = None
		jax_array = self._maybe_dequantize(jax_array, scales_jax_array)
		return jax_array.reshape(-1, head_dim)


def paged_flash_attention_kernel(
	lengths_ref,
	page_indices_ref,
	buffer_index_ref,
	step_ref,
	q_ref,
	k_pages_hbm_ref,
	k_scales_pages_hbm_ref,
	v_pages_hbm_ref,
	v_scales_pages_hbm_ref,
	o_ref,
	m_ref,
	l_ref,
	k_vmem_buffer,
	k_scales_vmem_buffer,
	v_vmem_buffer,
	v_scales_vmem_buffer,
	sem,
	*,
	batch_size: int,
	pages_per_compute_block: int,
	pages_per_sequence: int,
	mask_value: float,
	megacore_mode: str,
	program_ids=(),
):
	"""Pallas kernel for paged attention."""
	if program_ids:
		core_index, b, h, i = program_ids  # noqa
	else:
		core_index, b, h, i = (
			pl.program_id(0),
			pl.program_id(1),
			pl.program_id(2),
			pl.program_id(3),
		)
	num_kv_heads, _, page_size, _ = k_pages_hbm_ref.shape
	bk = page_size * pages_per_compute_block
	num_cores = pl.num_programs(0)

	b_step = num_cores if megacore_mode == "batch" else 1
	b_start = core_index if megacore_mode == "batch" else 0
	h_step = num_cores if megacore_mode == "kv_head" else 1
	h_start = core_index if megacore_mode == "kv_head" else 0

	h = h * h_step + h_start
	b = b * b_step + b_start
	length = lengths_ref[b]

	def compute_block_indices(b, h, i):
		def advance_b():
			next_b = b + b_step

			def advance_to_next_non_zero_length():
				next_next_b = next_b + b_step
				return lax.fori_loop(
					lax.div(next_next_b, b_step),
					lax.div(batch_size, b_step),
					lambda _, b: jnp.where(lengths_ref[b] == 0, b + b_step, b),
					next_next_b,
				)

			return (
				lax.cond(
					jnp.logical_and(next_b < batch_size, lengths_ref[next_b] == 0),
					advance_to_next_non_zero_length,
					lambda: next_b,
				),
				h_start,
				0,
			)

		def advance_h():
			next_h = h + h_step
			return lax.cond(next_h < num_kv_heads, lambda: (b, next_h, 0), advance_b)

		return lax.cond(i * bk < lengths_ref[b], lambda: (b, h, i), advance_h)

	def create_kv_async_copy_descriptors(b, h, i, buffer_index):
		page_offset = b * pages_per_sequence + i * pages_per_compute_block
		pages_to_load = pages_per_compute_block
		async_copy_k = MultiPageAsyncCopyDescriptor(
			k_pages_hbm_ref,
			k_scales_pages_hbm_ref,
			k_vmem_buffer.at[buffer_index],
			k_scales_vmem_buffer.at[buffer_index]
			if k_scales_vmem_buffer is not None
			else None,
			sem,
			page_indices_ref,
			page_offset,
			pages_to_load,
			h,
		)
		async_copy_v = MultiPageAsyncCopyDescriptor(
			v_pages_hbm_ref,
			v_scales_pages_hbm_ref,
			v_vmem_buffer.at[buffer_index],
			v_scales_vmem_buffer.at[buffer_index]
			if v_scales_vmem_buffer is not None
			else None,
			sem,
			page_indices_ref,
			page_offset,
			pages_to_load,
			h,
		)
		return async_copy_k, async_copy_v

	@pl.when(i * bk < length)
	def flash_attention():
		step = step_ref[0]
		buffer_index = buffer_index_ref[0]

		@pl.when(i == 0)
		def init():
			m_ref[...] = jnp.full_like(m_ref, -jnp.inf)
			l_ref[...] = jnp.zeros_like(l_ref)
			o_ref[...] = jnp.zeros_like(o_ref)

		@pl.when(step == 0)
		def prefetch_first_block():
			async_copy_k, async_copy_v = create_kv_async_copy_descriptors(
				b, h, i, buffer_index
			)
			async_copy_k.start()
			async_copy_v.start()

		next_b, next_h, next_i = compute_block_indices(b, h, i + 1)

		@pl.when(next_b < batch_size)
		def prefetch_next_block():
			next_buffer_index = jnp.where(buffer_index == 0, 1, 0)
			async_copy_next_k, async_copy_next_v = create_kv_async_copy_descriptors(
				next_b, next_h, next_i, next_buffer_index
			)
			async_copy_next_k.start()
			async_copy_next_v.start()
			buffer_index_ref[0] = next_buffer_index

		async_copy_k, async_copy_v = create_kv_async_copy_descriptors(b, h, i, buffer_index)
		q = q_ref[...].astype(jnp.float32)
		k = async_copy_k.wait_and_get_loaded()
		qk = jnp.einsum("hd,td->ht", q, k, preferred_element_type=jnp.float32)
		mask = i * bk + jax.lax.broadcasted_iota(jnp.int32, qk.shape, 1) < length
		qk = qk + jnp.where(mask, 0.0, mask_value)
		m_curr = qk.max(axis=-1)

		s_curr = jnp.exp(qk - m_curr[..., None])
		m_prev, l_prev = m_ref[...], l_ref[...]
		l_curr = jax.lax.broadcast_in_dim(s_curr.sum(axis=-1), l_prev.shape, (0,))
		m_curr = jax.lax.broadcast_in_dim(m_curr, m_prev.shape, (0,))
		m_next = jnp.maximum(m_prev, m_curr)
		alpha = jnp.exp(m_prev - m_next)
		beta = jnp.exp(m_curr - m_next)
		l_next = alpha * l_prev + beta * l_curr
		l_next_safe = jnp.where(l_next == 0.0, 1.0, l_next)

		v = async_copy_v.wait_and_get_loaded()
		o_curr_times_l_curr = jnp.dot(s_curr, v)

		m_ref[...], l_ref[...] = m_next, l_next_safe
		o_ref[...] = (
			(l_prev * alpha * o_ref[...] + beta * o_curr_times_l_curr) / l_next_safe
		).astype(o_ref.dtype)

		step_ref[0] = step + 1


def paged_flash_attention_kernel_inline_seq_dim(
	lengths_ref,
	page_indices_ref,
	buffer_index_ref,
	step_ref,
	q_ref,
	k_pages_hbm_ref,
	k_scales_pages_hbm_ref,
	v_pages_hbm_ref,
	v_scales_pages_hbm_ref,
	o_ref,
	m_ref,
	l_ref,
	k_vmem_buffer,
	k_scales_vmem_buffer,
	v_vmem_buffer,
	v_scales_vmem_buffer,
	sem,
	*,
	batch_size: int,
	pages_per_compute_block: int,
	pages_per_sequence: int,
	mask_value: float,
	megacore_mode: str,
):
	core_index, b, h = pl.program_id(0), pl.program_id(1), pl.program_id(2)

	# Initialize the output HBM buffers to avoid accessing garbage memory inside
	# the kernel body below.
	m_ref[...] = jnp.full_like(m_ref, -jnp.inf)
	l_ref[...] = jnp.zeros_like(l_ref)
	o_ref[...] = jnp.zeros_like(o_ref)

	def body(i, _):
		paged_flash_attention_kernel(
			lengths_ref,
			page_indices_ref,
			buffer_index_ref,
			step_ref,
			q_ref,
			k_pages_hbm_ref,
			k_scales_pages_hbm_ref,
			v_pages_hbm_ref,
			v_scales_pages_hbm_ref,
			o_ref,
			m_ref,
			l_ref,
			k_vmem_buffer,
			k_scales_vmem_buffer,
			v_vmem_buffer,
			v_scales_vmem_buffer,
			sem,
			batch_size=batch_size,
			pages_per_compute_block=pages_per_compute_block,
			pages_per_sequence=pages_per_sequence,
			mask_value=mask_value,
			megacore_mode=megacore_mode,
			program_ids=(core_index, b, h, i),
		)
		return ()

	bk = pages_per_compute_block * k_pages_hbm_ref.shape[-2]

	if megacore_mode == "batch":
		num_cores = pl.num_programs(0)
		length = lengths_ref[b * num_cores + core_index]
	else:
		length = lengths_ref[b]

	lax.fori_loop(0, lax.div(length + bk - 1, bk), body, ())


@functools.partial(
	jax.jit,
	static_argnames=[
		"pages_per_compute_block",
		"mask_value",
		"megacore_mode",
		"inline_seq_dim",
	],
)
def paged_attention(
	q: jax.Array,
	k_pages: Union[jax.Array, QuantizedTensor],
	v_pages: Union[jax.Array, QuantizedTensor],
	lengths: jax.Array,
	page_indices: jax.Array,
	*,
	mask_value: float = DEFAULT_MASK_VALUE,
	pages_per_compute_block: int,
	megacore_mode: Optional[str] = None,
	inline_seq_dim: bool = True,
) -> jax.Array:
	"""Paged grouped query attention.

	Args:
	  q: A [batch_size, num_heads, head_dim] jax.Array.
	  k_pages: A [num_kv_heads, total_num_pages, page_size, head_dim] jax.Array.
	  v_pages: A [num_kv_heads, total_num_pages, page_size, head_dim] jax.Array.
	  lengths: A i32[batch_size] jax.Array the length of each example.
	  page_indices: A i32[batch_size, pages_per_sequence] jax.Array. Each entry
	    should be in the range of [0, total_num_pages), indicating where to locate
	    the page in `k_pages` or `v_pages`.
	  mask_value: The value used for padding in attention. By default it is a very
	    negative floating point number.
	  pages_per_compute_block: how many pages to be processed in one flash
	    attention block in the pallas kernel.
	  megacore_mode: if set, enable megacore to parallelize the computation. Must
	    be one of ['kv_head', 'batch', None]. Caveat: set this only if megacore is
	    enabled, otherwise the kernel may hang. If you are not sure, leave it to
	    None.
	    * None: disable megacore parallelism.
	    * kv_head: megacore parallelism on KV heads; requires number of KV heads
	      divisible by 2.
	    * batch: megacore parallelism on batch dimension; requires batch divisible
	      by 2.
	  inline_seq_dim: whether to fuse kernel instances along the sequence dim into
	    one kernel.

	Returns:
	  The output of attention([batch_size, num_heads, head_dim]).
	"""
	if isinstance(k_pages, QuantizedTensor):
		k_pages, k_scales_pages = k_pages.weight, k_pages.scales
		assert isinstance(k_scales_pages, jax.Array)  # For typing.
		k_scales_pages = jnp.broadcast_to(
			k_scales_pages, (*k_scales_pages.shape[:-1], k_pages.shape[-1])
		)
	else:
		k_scales_pages = None
	if isinstance(v_pages, QuantizedTensor):
		v_pages, v_scales_pages = v_pages.weight, v_pages.scales
		assert isinstance(v_scales_pages, jax.Array)  # For typing.
		v_scales_pages = jnp.broadcast_to(
			v_scales_pages, (*v_scales_pages.shape[:-1], v_pages.shape[-1])
		)
	else:
		v_scales_pages = None

	batch_size, num_heads, head_dim = q.shape
	num_kv_heads, _, page_size, head_dim_k = k_pages.shape
	batch_size_paged_indices, pages_per_sequence = page_indices.shape

	if k_pages.shape != v_pages.shape:
		raise ValueError(
			f"k_pages and v_pages must have the same shape. Got {k_pages.shape} and"
			f" {v_pages.shape}"  # pytype: disable=attribute-error
		)
	if num_heads % num_kv_heads != 0:
		raise ValueError(
			"Number of Q heads must be divisible by number of KV heads. Got"
			f" {num_heads} and {num_kv_heads}."
		)
	if head_dim_k != head_dim:
		raise ValueError(
			"head_dim of Q must be the same as that of K/V. Got"
			f" {head_dim} and {head_dim_k}."
		)
	if pages_per_sequence % pages_per_compute_block != 0:
		raise ValueError(
			"pages_per_compute_block must be divisible by pages per sequence. Got"
			f" {pages_per_compute_block} and {pages_per_sequence}."
		)
	if lengths.shape != (batch_size,):
		raise ValueError("`lengths` and `q` must have the same batch size")
	if batch_size_paged_indices != batch_size:
		raise ValueError("`page_indices` and `q` must have the same batch size")
	if lengths.dtype != jnp.int32:
		raise ValueError("The dtype of `lengths` must be int32. Got {lengths.dtype}")

	if megacore_mode == "kv_head":
		if num_kv_heads % 2 != 0:
			raise ValueError(
				"number of KV heads must be even when megacore_mode is 'kv_head'"
			)
		num_cores = 2
	elif megacore_mode == "batch":
		if batch_size % 2 != 0:
			raise ValueError("batch size must be even when megacore_mode is 'batch'")
		num_cores = 2
	elif megacore_mode is None:
		num_cores = 1
	else:
		raise ValueError("megacore_mode must be one of ['kv_head', 'batch', None]")

	if (num_heads // num_kv_heads) % 8 != 0:
		# Reshape q to hint XLA to pick a <1x128> layout otherwise it will pick a
		# <8x128> layout for a <1x128> memref inside the kernel and error out.
		q = q.reshape(batch_size, num_heads, 1, head_dim)
		if megacore_mode == "kv_head":
			q_block_spec = pl.BlockSpec(
				(None, num_heads // num_kv_heads, None, head_dim),
				lambda core_index, b, h, *_: (b, h * num_cores + core_index, 0, 0),
			)
		elif megacore_mode == "batch":
			q_block_spec = pl.BlockSpec(
				(None, num_heads // num_kv_heads, None, head_dim),
				lambda core_index, b, h, *_: (b * num_cores + core_index, h, 0, 0),
			)
		else:
			q_block_spec = pl.BlockSpec(
				(None, num_heads // num_kv_heads, None, head_dim),
				lambda core_index, b, h, *_: (b, h, 0, 0),
			)
		q_dtype_for_kernel_launch = jnp.float32
	else:
		if megacore_mode == "kv_head":
			q_block_spec = pl.BlockSpec(
				(None, num_heads // num_kv_heads, head_dim),
				lambda core_index, b, h, *_: (b, h * num_cores + core_index, 0),
			)
		elif megacore_mode == "batch":
			q_block_spec = pl.BlockSpec(
				(None, num_heads // num_kv_heads, head_dim),
				lambda core_index, b, h, *_: (b * num_cores + core_index, h, 0),
			)
		else:
			q_block_spec = pl.BlockSpec(
				(None, num_heads // num_kv_heads, head_dim),
				lambda core_index, b, h, *_: (b, h, 0),
			)
		q_dtype_for_kernel_launch = q.dtype

	if inline_seq_dim:
		kernel = paged_flash_attention_kernel_inline_seq_dim
		grid = (
			num_cores,
			batch_size // num_cores if megacore_mode == "batch" else batch_size,
			num_kv_heads // num_cores if megacore_mode == "kv_head" else num_kv_heads,
		)
		dimension_sematics = ("parallel", "arbitrary", "arbitrary")
	else:
		kernel = paged_flash_attention_kernel
		grid = (
			num_cores,
			batch_size // num_cores if megacore_mode == "batch" else batch_size,
			num_kv_heads // num_cores if megacore_mode == "kv_head" else num_kv_heads,
			pages_per_sequence // pages_per_compute_block,
		)  # type: ignore
		dimension_sematics = ("parallel", "arbitrary", "arbitrary", "arbitrary")  # type: ignore

	if k_scales_pages is not None and v_scales_pages is not None:
		in_specs = [
			q_block_spec,
			pl.BlockSpec(memory_space=pltpu.TPUMemorySpace.ANY),
			pl.BlockSpec(memory_space=pltpu.TPUMemorySpace.ANY),
			pl.BlockSpec(memory_space=pltpu.TPUMemorySpace.ANY),
			pl.BlockSpec(memory_space=pltpu.TPUMemorySpace.ANY),
		]
		scratch_shapes = (
			pltpu.VMEM(
				(
					2,  # For double buffering during DMA copies.
					pages_per_compute_block,
					page_size,
					head_dim,
				),
				k_pages.dtype,
			),  # k_pages buffer
			pltpu.VMEM(
				(
					2,  # For double buffering during DMA copies.
					pages_per_compute_block,
					page_size,
					head_dim,
				),
				k_scales_pages.dtype,  # pytype: disable=attribute-error
			),  # k_scales_pages buffer
			pltpu.VMEM(
				(
					2,  # For double buffering during DMA copies.
					pages_per_compute_block,
					page_size,
					head_dim,
				),
				v_pages.dtype,
			),  # v_pages buffer
			pltpu.VMEM(
				(
					2,  # For double buffering during DMA copies.
					pages_per_compute_block,
					page_size,
					head_dim,
				),
				v_scales_pages.dtype,  # pytype: disable=attribute-error
			),  # v_scales_pages buffer
			pltpu.SemaphoreType.DMA,
		)
	else:
		in_specs = [
			q_block_spec,
			pl.BlockSpec(memory_space=pltpu.TPUMemorySpace.ANY),
			None,  # type: ignore[list-item]
			pl.BlockSpec(memory_space=pltpu.TPUMemorySpace.ANY),
			None,  # type: ignore[list-item]
		]
		scratch_shapes = (
			pltpu.VMEM(
				(
					2,  # For double buffering during DMA copies.
					pages_per_compute_block,
					page_size,
					head_dim,
				),
				k_pages.dtype,
			),  # k_pages buffer
			None,
			pltpu.VMEM(
				(
					2,  # For double buffering during DMA copies.
					pages_per_compute_block,
					page_size,
					head_dim,
				),
				v_pages.dtype,
			),  # v_pages buffer
			None,
			pltpu.SemaphoreType.DMA,
		)

	out, _, _ = pl.pallas_call(
		functools.partial(
			kernel,
			pages_per_sequence=pages_per_sequence,
			batch_size=batch_size,
			pages_per_compute_block=pages_per_compute_block,
			mask_value=mask_value,
			megacore_mode=megacore_mode,
		),
		grid_spec=pltpu.PrefetchScalarGridSpec(
			# There are 4 scalars prefetched per kernel call: `lengths_ref`,
			# `page_indices_ref`, `buffer_index_ref`, `step_ref`
			num_scalar_prefetch=4,
			in_specs=in_specs,
			out_specs=[
				q_block_spec,
				q_block_spec,
				q_block_spec,
			],
			grid=grid,
			scratch_shapes=scratch_shapes,
		),
		compiler_params=dict(mosaic=dict(dimension_semantics=dimension_sematics)),
		out_shape=[
			jax.ShapeDtypeStruct(q.shape, q_dtype_for_kernel_launch),
			jax.ShapeDtypeStruct((*q.shape[:-1], 1), jnp.float32),
			jax.ShapeDtypeStruct((*q.shape[:-1], 1), jnp.float32),
		],
	)(
		lengths,
		page_indices.reshape(-1),
		jnp.zeros((1,), jnp.int32),  # buffer index
		jnp.zeros((1,), jnp.int32),  # step
		q.astype(q_dtype_for_kernel_launch),
		k_pages,
		k_scales_pages,
		v_pages,
		v_scales_pages,
	)
	return out.reshape(batch_size, num_heads, head_dim).astype(q.dtype)



from fjformer.pallas_operations import efficient_attention as efficient_attention
from fjformer.pallas_operations import gpu as gpu
from fjformer.pallas_operations import pallas_attention as pallas_attention
from fjformer.pallas_operations import tpu as tpu
from fjformer.pallas_operations.jax_flash_attention2 import flash_attention2 as flash_attention2


from __future__ import annotations

from collections.abc import Mapping
import dataclasses
import enum
import functools
from typing import Any, Callable, Literal, NamedTuple, Optional, Union, overload

import jax
from jax import ad_checkpoint
from jax import lax
from jax import tree_util
from jax.experimental import pallas as pl
from jax.experimental.pallas import tpu as pltpu
from fjformer.pallas_operations.tpu.splash_attention import (
	splash_attention_mask as mask_lib,
)
from fjformer.pallas_operations.tpu.splash_attention import (
	splash_attention_mask_info as mask_info_lib,
)
import jax.numpy as jnp
import numpy as np

partial = functools.partial
DEFAULT_MASK_VALUE = -0.7 * float(np.finfo(np.dtype("float32")).max)
NUM_LANES = 128
NUM_SUBLANES = 8
# We predefine some useful dimension numbers for dot_general
NN_DIM_NUMBERS = (((1,), (0,)), ((), ()))  # standard matmul
NT_DIM_NUMBERS = (((1,), (1,)), ((), ()))  # RHS transposed


# mypy: ignore-errors


class SegmentIds(NamedTuple):
	"""SegmentIds for Q and KV sequences.

	SegmentIds are a mechanims to ensure that there is no cross-attention between
	segments (fraction of a sequence) that have been concatenated together into a
	sequence. Each array is a list of ids (integers). Only tokens with the same
	id are allowed to attend to each other.

	The static mask (e.g. causal) is "and-ed" with the segment id mask to form
	the actual attention mask. It is important that the latter does not have any
	all-zero rows (along dimension kv). Otherwise it would result in a invalid
	softmax (the denominator would be 0).
	This condition holds for causal self-attention because in this case segment
	ids form a block diagonal matrix so at least one element in each row is set.
	It is easy to break this condition with non-self-attention configurations.
	Attributes:
	  q: segment ids along the Q sequence
	  kv: segment ids along the KV sequence
	"""

	q: jax.Array  # [q_seq_len]
	kv: jax.Array  # [kv_seq_len]


# Return type of SplashAttention function that implements the custom vjp rule.
SplashCustomReturnType = Union[
	# out, no residuals
	jax.Array,
	# out, residuals:
	tuple[jax.Array, tuple[jax.Array,]],
]

SplashResidualsType = tuple[
	jax.Array,  # q
	jax.Array,  # k
	jax.Array,  # v
	Optional[SegmentIds],  # segment_ids
	jax.Array,  # out
	jax.Array,  # logsumexp
	Optional[mask_info_lib.MaskInfo],  # dq_mask_info
	Optional[mask_info_lib.MaskInfo],  # dkv_mask_info
]

MaskFunctionType = Callable[..., jax.Array]


def get_kernel_name(
	block_metadata: Mapping[str, Any],
	is_mqa: bool,
	save_residuals: bool,
	is_segmented: bool,
	phase: str,
) -> str:
	"""Returns a unique name for all SplashAttention kernel variants."""
	assert phase == "dq" or phase == "dkv" or phase == "fwd"
	# Saving residuals is supported only for the fwd phase.
	assert not save_residuals or phase == "fwd"
	residuals = ""
	if save_residuals:
		residuals = "_residuals"
	elif phase == "fwd":
		residuals = "_no_residuals"
	attention_type = "mqa" if is_mqa else "mha"
	segments = "_segmented" if is_segmented else ""
	return f"splash_{attention_type}_{phase}{segments}{residuals}_" + "_".join(
		f"{k}={v}" for k, v in sorted(block_metadata.items())
	)


# Reference attention implementations


@overload
def _attention_reference(
	mask: jax.Array,
	q: jax.Array,
	k: jax.Array,
	v: jax.Array,
	segment_ids: SegmentIds | None,
	save_residuals: Literal[False],
	mask_value: float,
	custom_type: str,
	attn_logits_soft_cap: float | None,
) -> jax.Array: ...


@overload
def _attention_reference(
	mask: jax.Array,
	q: jax.Array,
	k: jax.Array,
	v: jax.Array,
	segment_ids: SegmentIds | None,
	save_residuals: Literal[True],
	mask_value: float,
	custom_type: str,
	attn_logits_soft_cap: float | None,
) -> tuple[jax.Array, tuple[jax.Array]]: ...


def _attention_reference(
	mask: jax.Array,  # [q_seq_len, kv_seq_len]
	q: jax.Array,  # [q_seq_len, head_dim]
	k: jax.Array,  # [kv_seq_len, head_dim]
	v: jax.Array,  # [kv_seq_len, head_dim]
	segment_ids: SegmentIds | None,
	mask_value: float,
	save_residuals: bool,
	custom_type: str,
	attn_logits_soft_cap: float | None,
):
	return _attention_reference_default(  # pytype: disable=bad-return-type
		mask,
		q,
		k,
		v,
		segment_ids,
		mask_value,
		save_residuals,
		custom_type,
		attn_logits_soft_cap,
	)


def _attention_reference_default(
	mask: jax.Array,  # [q_seq_len, kv_seq_len]
	q: jax.Array,  # [q_seq_len, head_dim]
	k: jax.Array,  # [kv_seq_len, head_dim]
	v: jax.Array,  # [kv_seq_len, head_dim]
	segment_ids: SegmentIds | None,
	mask_value: float,
	save_residuals: bool,
	custom_type: str,
	attn_logits_soft_cap: float | None,
):
	del custom_type
	logits = jnp.einsum("sd,td->st", q.astype(jnp.float32), k.astype(jnp.float32))

	if segment_ids is not None:
		mask = jnp.logical_and(mask, segment_ids.q[:, None] == segment_ids.kv[None, :])

	if attn_logits_soft_cap is not None:
		logits = jnp.tanh(logits / attn_logits_soft_cap)
		logits = logits * attn_logits_soft_cap

	logits = jnp.where(mask, logits, mask_value)
	m = logits.max(axis=-1)
	s = jnp.exp(logits - m[..., None])
	l = s.sum(axis=-1)
	s = s / l[..., None]

	o = jnp.einsum("st,td->sd", s, v.astype(jnp.float32))

	logsumexp = m + jnp.log(l)
	if save_residuals:
		return o, (logsumexp,)
	return o


def attention_reference(
	mask: jax.Array,  # [q_seq_len, kv_seq_len]
	q: jax.Array,  # [q_seq_len, head_dim]
	k: jax.Array,  # [kv_seq_len, head_dim]
	v: jax.Array,  # [kv_seq_len, head_dim]
	segment_ids: SegmentIds | None,
	*,
	mask_value: float = DEFAULT_MASK_VALUE,
	save_residuals: bool = False,
	custom_type: str = "flash",
	attn_logits_soft_cap: float | None = None,
) -> SplashCustomReturnType:
	return _attention_reference(  # pytype: disable=wrong-arg-types
		mask,
		q,
		k,
		v,
		segment_ids,
		mask_value=mask_value,
		save_residuals=save_residuals,
		custom_type=custom_type,
		attn_logits_soft_cap=attn_logits_soft_cap,
	)


def _attention_reference_custom_fwd(
	mask: jax.Array,  # [q_seq_len, kv_seq_len]
	q: jax.Array,  # [q_seq_len, head_dim]
	k: jax.Array,  # [kv_seq_len, head_dim]
	v: jax.Array,  # [kv_seq_len, head_dim]
	segment_ids: SegmentIds | None,
	mask_value: float,
	save_residuals: bool,
	custom_type: str,
	attn_logits_soft_cap: float | None,
):
	if save_residuals:
		raise NotImplementedError("Higher-order AD not supported.")

	o, (logsumexp,) = _attention_reference(
		mask,
		q,
		k,
		v,
		segment_ids,
		mask_value=mask_value,
		save_residuals=True,
		custom_type=custom_type,
		attn_logits_soft_cap=attn_logits_soft_cap,
	)
	return o, (mask, q, k, v, segment_ids, o, logsumexp)


def _attention_reference_custom_bwd(
	mask_value: float,
	save_residuals: bool,
	custom_type: str,
	attn_logits_soft_cap: float | None,
	res,
	do: jax.Array,
) -> tuple[None, jax.Array, jax.Array, jax.Array, None]:
	del save_residuals
	mask, q, k, v, segment_ids, o, logsumexp = res

	uncapped_logits = jnp.einsum("qc,kc->qk", q, k, preferred_element_type=jnp.float32)

	if attn_logits_soft_cap is not None:
		logits = jnp.tanh(uncapped_logits / attn_logits_soft_cap)
		logits = logits * attn_logits_soft_cap
	else:
		logits = uncapped_logits

	if segment_ids is not None:
		mask = jnp.logical_and(mask, segment_ids.q[:, None] == segment_ids.kv[None, :])
	logits = jnp.where(mask, logits, mask_value)

	p = jnp.exp(logits - logsumexp[..., None])
	do = do.astype(jnp.float32)  # pytype: disable=attribute-error
	dv = jnp.einsum("pt,pd->td", p, do).astype(v.dtype)
	dp = jnp.einsum("pd,td->pt", do, v.astype(jnp.float32))

	# These two ways of computing ds are mathematically equivalent. The first
	# involves reducing over the head_dim dimension and the second involves
	# reducing over a sequence dimension. They tend to produce slightly different
	# numerics.
	if custom_type == "flash":
		di = jnp.sum(o.astype(jnp.float32) * do, axis=-1)[..., None]
	else:
		di = jnp.einsum("st,st->s", dp, p)[:, None]
	ds = (dp - di) * p
	if attn_logits_soft_cap is not None:
		normalized = uncapped_logits / attn_logits_soft_cap
		d = jnp.tanh(normalized)
		g = ds * (1 - d)
		ds = g + g * d
	dk = jnp.einsum("sd,st->td", q.astype(jnp.float32), ds).astype(k.dtype)
	dq = jnp.einsum("st,td->sd", ds, k.astype(jnp.float32)).astype(q.dtype)
	return None, dq, dk, dv, None


_attention_reference_custom = jax.custom_vjp(
	_attention_reference, nondiff_argnums=(5, 6, 7, 8)
)
_attention_reference_custom.defvjp(
	_attention_reference_custom_fwd, _attention_reference_custom_bwd
)


def attention_reference_custom(
	mask: jax.Array,  # [q_seq_len, kv_seq_len]
	q: jax.Array,  # [q_seq_len, head_dim]
	k: jax.Array,  # [kv_seq_len, head_dim]
	v: jax.Array,  # [kv_seq_len, head_dim]
	segment_ids: SegmentIds | None,
	*,
	mask_value: float = DEFAULT_MASK_VALUE,
	save_residuals: bool = False,
	custom_type: str = "flash",
	attn_logits_soft_cap: float | None = None,
):
	return _attention_reference_custom(
		mask,
		q,
		k,
		v,
		segment_ids,
		mask_value,
		save_residuals,
		custom_type=custom_type,
		attn_logits_soft_cap=attn_logits_soft_cap,
	)


def make_attention_reference(
	mask: mask_lib.Mask | np.ndarray,
	is_mqa: bool,
	backward_impl: str = "vanilla",
	**params: Any,
) -> Callable:
	@partial(
		jax.jit,
		static_argnames=[
			"mask_value",
			"save_residuals",
			"attn_logits_soft_cap",
		],
	)
	def _wrapped(
		mask: jax.Array,
		q: jax.Array,
		k: jax.Array,
		v: jax.Array,
		segment_ids: SegmentIds | None = None,
		*,
		mask_value: float = DEFAULT_MASK_VALUE,
		save_residuals: bool = False,
		attn_logits_soft_cap: float | None = None,
	):
		if backward_impl == "custom":
			attn_impl = partial(
				attention_reference_custom,
				custom_type="flash",
			)
		elif backward_impl == "custom_vanilla":
			attn_impl = partial(
				attention_reference_custom,
				custom_type="vanilla",
			)
		else:
			attn_impl = attention_reference
		func = partial(
			attn_impl,
			mask_value=mask_value,
			save_residuals=save_residuals,
			attn_logits_soft_cap=attn_logits_soft_cap,
			**params,
		)

		if is_mqa:
			func = jax.vmap(func, in_axes=(0, 0, None, None, None))
			is_grouped = False
		else:
			# In grouped attention (1 < num_kv_heads && num_kv_heads < num_q_heads).
			# We interleave the KV heads across the Q heads.
			# For example: for 8 Q heads and 4 KV heads:
			# Q head [0, 1] see KV head 0
			# Q head [2, 3] see KV head 1
			# Q head [4, 5] see KV head 2
			# Q head [6, 7] see KV head 3
			#
			# The following implementation reshapes Q to expose KV heads and vmaps
			# Across the Q heads so it is similar to MQA.
			# Alternatively we can replicate K/V to match Q like so:
			# k = jnp.repeat(k, q_heads_per_kv_head, axis=0)
			# v = jnp.repeat(v, q_heads_per_kv_head, axis=0)

			kv_heads = k.shape[0]
			assert kv_heads == v.shape[0]
			q_heads, q_seq_len, head_dim = q.shape
			is_grouped = kv_heads < q_heads
			if is_grouped:
				assert q_heads % kv_heads == 0
				assert mask.shape[0] == q_heads
				q_heads_per_kv_head = q_heads // kv_heads
				q = q.reshape((kv_heads, q_heads_per_kv_head, q_seq_len, head_dim))
				mask = mask.reshape((kv_heads, q_heads_per_kv_head, *mask.shape[1:]))

				# Inner-most vmap: iterate over the q heads.
				func = jax.vmap(func, in_axes=(0, 0, None, None, None))

			# Outer-most vmap: iterate over the kv heads.
			func = jax.vmap(func, in_axes=(0, 0, 0, 0, None))

		out = func(mask, q, k, v, segment_ids)

		if is_grouped:

			def reshape_activations(activations):
				if activations.ndim == 4:  # pytype: disable=attribute-error
					kv_heads, q_heads_per_kv_head, q_seq_len, head_dim = (
						activations.shape
					)  # pytype: disable=attribute-error
					return activations.reshape(  # pytype: disable=attribute-error
						kv_heads * q_heads_per_kv_head, q_seq_len, head_dim
					)
				return activations

			def reshape_residuals(residuals):
				if residuals.ndim == 3:
					kv_heads, q_heads_per_kv_head, q_seq_len = residuals.shape
					return residuals.reshape(kv_heads * q_heads_per_kv_head, q_seq_len)
				return residuals

			if save_residuals:
				assert isinstance(out, tuple)
				assert isinstance(out[1], tuple)

				return (reshape_activations(out[0]), (reshape_residuals(out[1][0]),))
			else:
				return reshape_activations(out)
		else:
			return out

	return functools.partial(_wrapped, jnp.array(mask[:, :, :]))


make_masked_mha_reference = partial(make_attention_reference, is_mqa=False)
make_masked_mqa_reference = partial(make_attention_reference, is_mqa=True)


# Splash attention implementation


# We use an IntEnum to make it JSON serializable as regen metadata.
class QKVLayout(enum.IntEnum):
	HEAD_DIM_MINOR = enum.auto()  # [..., seq_len, head_dim]
	SEQ_MINOR = enum.auto()  # [..., head_dim, seq_len]


def from_head_minor(vals: tuple[Any, ...], layout: QKVLayout):
	if layout == QKVLayout.HEAD_DIM_MINOR:
		return vals
	return (*vals[:-2], vals[-1], vals[-2])


@dataclasses.dataclass(unsafe_hash=True)
class BlockSizes:
	"""Tile sizes parameterizing SplashAttention kernels.

	Those parameters have negligible effect on numerics, but affect performance
	greatly.

	Note that changing the layouts only influences the physical layout that the
	kernel will enforce. The logical interface to splash attention always takes
	the head dimension as the minormost one.
	"""

	block_q: int
	block_kv: int
	block_kv_compute: int | None = None

	block_q_dkv: int | None = None
	block_kv_dkv: int | None = None
	block_kv_dkv_compute: int | None = None

	block_q_dq: int | None = None
	block_kv_dq: int | None = None

	use_fused_bwd_kernel: bool = False

	q_layout: QKVLayout = QKVLayout.HEAD_DIM_MINOR
	k_layout: QKVLayout = QKVLayout.HEAD_DIM_MINOR
	v_layout: QKVLayout = QKVLayout.HEAD_DIM_MINOR

	def __post_init__(self):
		if self.block_kv_compute is None:
			self.block_kv_compute = self.block_kv
		if self.block_kv_dkv_compute is None:
			self.block_kv_dkv_compute = self.block_kv_dkv
		if self.use_fused_bwd_kernel:
			if self.block_q_dq is not None or self.block_kv_dq is not None:
				raise ValueError(
					"Block sizes for dq kernel are not needed with a fused kernel."
				)

	@property
	def has_backward_blocks(self) -> bool:
		backward_blocks = (
			self.block_q_dkv,
			self.block_kv_dkv,
			self.block_kv_dkv_compute,
		)
		if not self.use_fused_bwd_kernel:
			backward_blocks += (self.block_q_dq, self.block_kv_dq)
		return all(b is not None for b in backward_blocks)

	@classmethod
	def get_default(cls):
		return BlockSizes(
			block_q=128,
			block_kv=128,
			block_kv_compute=128,
			block_q_dkv=128,
			block_kv_dkv=128,
			block_kv_dkv_compute=128,
			block_q_dq=128,
			block_kv_dq=128,
		)


def _next_nonzero(
	h,
	i,
	j,
	data_next_ref,
	block_mask_ref,
	m_next_ref,
	next_i=False,
):
	assert (data_next_ref is None) == (block_mask_ref is None)

	if data_next_ref is None and block_mask_ref is None:
		# Handle the case in which we have no masking nor next data information.
		# Simply fetch the next data and apply the mask for every block.
		assert m_next_ref is None
		next_data = i if next_i else j
		return (
			next_data,
			None,  # next mask
			True,  # should run
			False,  # should not mask
		)

	assert data_next_ref.shape == block_mask_ref.shape
	assert m_next_ref is None or data_next_ref.shape[0] == m_next_ref.shape[0]

	# We are working with one head only. Force the head index to 0.
	if data_next_ref.shape[0] == 1:
		h = 0

	# When scalar-memory data is of types smaller than int32, then we have to
	# upcast it back to use it in the kernel.

	to_i32 = lambda x: x.astype(jnp.int32)

	is_nonzero = to_i32(block_mask_ref[h, i, j]) > 0
	if m_next_ref is None:
		should_not_mask = True
		next_m = None
	else:
		should_not_mask = to_i32(block_mask_ref[h, i, j]) != 1
		next_m = to_i32(m_next_ref[h, i, j])
	next_j = to_i32(data_next_ref[h, i, j])
	return next_j, next_m, is_nonzero, should_not_mask


def _apply_mask_and_soft_cap(
	qk: jax.Array,
	mask_value: float,
	should_not_mask,
	mask_ref,
	q_sequence_ref,
	q_segment_ids_ref,
	kv_segment_ids_ref,
	*,
	attn_logits_soft_cap: float,
	k_slice: pl.Slice,
	k_offset: int | jax.Array,
	bq: int,
	k_in_lanes=True,
	mask_function=None,
) -> jax.Array | tuple[jax.Array, jax.Array, jax.Array, jax.Array]:
	assert mask_ref is None or q_sequence_ref is None
	assert (q_sequence_ref is None) == (mask_function is None)

	masks = []
	if mask_ref is not None:
		if k_in_lanes:
			mask = pl.load(mask_ref, (slice(None), k_slice))
		else:
			mask = pl.load(mask_ref, (k_slice, slice(None)))

		snm = jnp.where(should_not_mask, 1, 0)
		masks.append(jnp.bitwise_or(mask, jnp.broadcast_to(snm, mask.shape)))

	if mask_function is not None:
		# Compute the mask using the given q_sequence indices.
		# KV indices are computed on the fly. This works because we only support Q
		# sequence sharding. If we wanted to compute Q indices too, then we would
		# need to keep into account the current shard along Q sequence.

		if k_in_lanes:
			assert q_sequence_ref.shape == (bq, NUM_LANES)

			k_sequence = k_offset + jax.lax.broadcasted_iota(jnp.int32, (bq, k_slice.size), 1)

			repeats, rem = divmod(k_slice.size, NUM_LANES)
			assert rem == 0
			q_sequence = pltpu.repeat(
				q_sequence_ref[...], repeats, axis=1
			)  # [bq, k_slice.size]
		else:
			assert q_sequence_ref.shape == (NUM_SUBLANES, bq)

			k_sequence = k_offset + jax.lax.broadcasted_iota(jnp.int32, (k_slice.size, bq), 0)
			q_sequence = pl.load(q_sequence_ref, (pl.ds(1), slice(None)))  # [1, bq]
			q_sequence = jnp.broadcast_to(q_sequence, (k_slice.size, bq))

		assert q_sequence.shape == k_sequence.shape
		masks.append(
			mask_function(q_sequence, k_sequence)
		)  # pytype: disable=wrong-arg-count

	if q_segment_ids_ref is not None:
		if k_in_lanes:
			kv_ids = pl.load(kv_segment_ids_ref, (pl.ds(1), k_slice))  # [1, k_slice]
			repeats, rem = divmod(kv_ids.shape[1], NUM_LANES)
			if rem:
				raise NotImplementedError(f"block_kv must be a multiple of {NUM_LANES}")
			q_ids = pltpu.repeat(q_segment_ids_ref[:], repeats, axis=1)  # [bq, bkv]
		else:
			assert bq == q_segment_ids_ref.shape[-1]
			repeats, rem = divmod(bq, NUM_LANES)
			if rem:
				raise NotImplementedError(f"block_q must be a multiple of {NUM_LANES}")
			kv_ids = pltpu.repeat(
				pl.load(kv_segment_ids_ref, (k_slice, slice(None))), repeats, axis=1
			)  # [k_slice, bq]
			q_ids = pl.load(q_segment_ids_ref, (pl.ds(1), slice(None)))  # [1, bq]
		masks.append(q_ids == kv_ids)

	def cap_logits(logits):
		if attn_logits_soft_cap is not None:
			logits = jnp.tanh(qk / attn_logits_soft_cap)
			return logits * attn_logits_soft_cap
		else:
			return logits

	if masks:
		mask = functools.reduce(jnp.logical_and, masks)
		qk = cap_logits(qk)
		qk = jnp.where(mask, qk, mask_value)
	else:
		qk = cap_logits(qk)
	return qk


def flash_attention_kernel(
	# Prefetched inputs
	data_next_ref,
	block_mask_ref,
	mask_next_ref,
	# Inputs
	q_ref,
	k_ref,
	v_ref,
	q_segment_ids_ref,
	kv_segment_ids_ref,
	mask_ref,
	q_sequence_ref,
	# Outputs
	m_scratch_ref,
	l_scratch_ref,
	o_scratch_ref,
	o_ref,
	logsumexp_ref=None,
	*,
	mask_value: float,
	grid_width: int,
	bq: int,
	bkv: int,
	bkv_compute: int,
	head_dim: int,
	q_layout: QKVLayout,
	k_layout: QKVLayout,
	v_layout: QKVLayout,
	attn_logits_soft_cap: float | None,
	mask_function: MaskFunctionType | None,
):
	float32 = jnp.float32
	HEAD_DIM_MINOR = QKVLayout.HEAD_DIM_MINOR

	head_dim_repeats, rem = divmod(head_dim, NUM_LANES)
	if rem != 0:
		raise NotImplementedError(f"{head_dim=} should be a multiple of {NUM_LANES}")

	h, i, j = pl.program_id(0), pl.program_id(1), pl.program_id(2)

	@pl.when(j == 0)
	def init():
		o_scratch_ref[...] = jnp.zeros_like(o_scratch_ref)
		m_scratch_ref[...] = jnp.full_like(m_scratch_ref, mask_value)
		l_scratch_ref[...] = jnp.zeros_like(l_scratch_ref)

	_, _, should_run, should_not_mask = _next_nonzero(
		h,
		i,
		j,
		data_next_ref,
		block_mask_ref,
		mask_next_ref,
	)

	def body(kv_compute_index, _):
		slice_k = pl.ds(kv_compute_index * bkv_compute, bkv_compute)
		m_prev, l_prev = m_scratch_ref[...], l_scratch_ref[...]
		assert m_prev.shape == (bq, NUM_LANES)
		assert l_prev.shape == (bq, NUM_LANES)

		q = q_ref[...] if q_layout == HEAD_DIM_MINOR else q_ref[...].T
		qk_dims = NT_DIM_NUMBERS if k_layout == HEAD_DIM_MINOR else NN_DIM_NUMBERS
		if k_layout == HEAD_DIM_MINOR:
			k = pl.load(k_ref, (slice_k, slice(None)))
		else:
			k = pl.load(k_ref, (slice(None), slice_k))
		qk = lax.dot_general(q, k, qk_dims, preferred_element_type=float32)

		assert qk.shape == (bq, bkv_compute)
		apply_mask_and_soft_cap = functools.partial(
			_apply_mask_and_soft_cap,
			qk,
			mask_value,
			should_not_mask,
			mask_ref,
			q_sequence_ref,
			q_segment_ids_ref,
			kv_segment_ids_ref,
			attn_logits_soft_cap=attn_logits_soft_cap,
			k_slice=slice_k,
			k_offset=j * bkv + kv_compute_index * bkv_compute,
			bq=bq,
			mask_function=mask_function,
		)

		qk = apply_mask_and_soft_cap()

		m_curr = qk.max(axis=-1)[:, None]  # pytype: disable=attribute-error
		assert m_curr.shape == (bq, 1)
		m_next = jnp.maximum(m_prev, m_curr)
		assert m_next.shape == (bq, NUM_LANES)

		bkv_repeats, rem = divmod(bkv_compute, NUM_LANES)
		if rem != 0:
			raise NotImplementedError(f"{bkv_compute=} should be a multiple of {NUM_LANES}")

		s_curr = jnp.exp(qk - pltpu.repeat(m_next, bkv_repeats, axis=1))
		assert s_curr.shape == (bq, bkv_compute)

		l_curr = jax.lax.broadcast_in_dim(s_curr.sum(axis=-1), l_prev.shape, (0,))
		assert l_curr.shape == (bq, NUM_LANES)

		alpha = jnp.exp(m_prev - m_next)
		l_next = l_curr + alpha * l_prev
		m_scratch_ref[...], l_scratch_ref[...] = m_next, l_next

		sv_dims = NN_DIM_NUMBERS if v_layout == HEAD_DIM_MINOR else NT_DIM_NUMBERS
		if v_layout == HEAD_DIM_MINOR:
			v = pl.load(v_ref, (slice_k, slice(None)))
		else:
			v = pl.load(v_ref, (slice(None), slice_k))
		v = v.astype(float32)
		o_curr = lax.dot_general(s_curr, v, sv_dims)

		alpha_o = pltpu.repeat(alpha, head_dim_repeats, axis=1)
		o_scratch_ref[:] = alpha_o * o_scratch_ref[:] + o_curr

	@pl.when(should_run)
	def run():
		assert bkv % bkv_compute == 0
		num_iters = k_ref.shape[0 if k_layout == HEAD_DIM_MINOR else 1] // bkv_compute
		lax.fori_loop(0, num_iters, body, None, unroll=True)

	@pl.when(j == grid_width - 1)
	def end():
		l = l_scratch_ref[...]
		l_inv = pltpu.repeat(1.0 / l, head_dim_repeats, axis=1)
		o_ref[...] = (o_scratch_ref[...] * l_inv).astype(o_ref.dtype)
		if logsumexp_ref is not None:
			assert logsumexp_ref.shape == (bq, NUM_LANES)
			logsumexp_ref[...] = (jnp.log(l) + m_scratch_ref[...]).astype(logsumexp_ref.dtype)

		m_scratch_ref[...] = jnp.zeros_like(m_scratch_ref)
		l_scratch_ref[...] = jnp.zeros_like(l_scratch_ref)
		o_scratch_ref[...] = jnp.zeros_like(o_scratch_ref)


@overload
def _splash_attention_forward(
	fwd_mask_info: mask_info_lib.MaskInfo,
	q: jax.Array,
	k: jax.Array,
	v: jax.Array,
	segment_ids: SegmentIds | None,
	mask_value: float,
	is_mqa: bool,
	block_sizes: BlockSizes,
	residual_checkpoint_name: str | None,
	mask_function: MaskFunctionType | None,
	save_residuals: Literal[False] = False,
	attn_logits_soft_cap: float | None = None,
) -> jax.Array: ...


@overload
def _splash_attention_forward(
	fwd_mask_info: mask_info_lib.MaskInfo,
	q: jax.Array,
	k: jax.Array,
	v: jax.Array,
	segment_ids: SegmentIds | None,
	mask_value: float,
	is_mqa: bool,
	block_sizes: BlockSizes,
	residual_checkpoint_name: str | None,
	mask_function: MaskFunctionType | None,
	save_residuals: Literal[True],
	attn_logits_soft_cap: float | None = None,
) -> SplashCustomReturnType: ...


def _div(dividend: int, divisor: int):
	if divisor == 1:
		return dividend

	return lax.div(dividend, divisor)


def _splash_attention_forward(
	fwd_mask_info: mask_info_lib.MaskInfo,
	q: jax.Array,
	k: jax.Array,
	v: jax.Array,
	segment_ids: SegmentIds | None,
	mask_value: float,
	is_mqa: bool,
	block_sizes: BlockSizes,
	residual_checkpoint_name: str | None,
	save_residuals: bool,
	mask_function: MaskFunctionType | None,
	attn_logits_soft_cap: float | None = None,
) -> SplashCustomReturnType:
	num_q_heads, q_seq_len, head_dim = q.shape
	bq, bkv = block_sizes.block_q, block_sizes.block_kv
	bkv_compute = block_sizes.block_kv_compute

	if is_mqa:
		expected_kv_rank = 2
		kv_head_dimension = 1
		kv_seq_len_dimension = 0
		num_kv_heads = 1
	else:
		expected_kv_rank = 3
		kv_head_dimension = 2
		kv_seq_len_dimension = 1
		num_kv_heads = k.shape[0]

	if len(k.shape) != expected_kv_rank:
		raise ValueError(
			f"Expected {expected_kv_rank}-dim 'key' tensor for MQA. Instead got a"
			f" {len(k.shape)}-dim one."
		)

	if k.shape[kv_head_dimension] != head_dim:
		raise ValueError(
			f"Expected 'key' head dimension to be: {head_dim}. Instead got:"
			f" {k.shape[kv_head_dimension]}."
		)

	if not is_mqa and num_q_heads % num_kv_heads != 0:
		raise ValueError(
			f"In MHA, expected number of 'key' heads ({num_kv_heads}) to be a"
			f" multiple of the number of 'query' heads ({num_q_heads})"
		)

	if k.shape != v.shape:
		raise ValueError(
			f"Expected 'key' {k.shape} and 'value' {v.shape} to have the same" " shape."
		)

	if bkv % bkv_compute:
		raise ValueError(f"{bkv=} must be a multiple of {bkv_compute=}.")
	if bkv_compute % NUM_LANES:
		raise ValueError(f"{bkv_compute=} must be a multiple of {NUM_LANES}.")

	kv_seq_len = k.shape[kv_seq_len_dimension]

	q_heads_per_kv_head = num_q_heads // num_kv_heads

	if segment_ids is not None:
		if segment_ids.q.shape != (q_seq_len,):
			raise ValueError(
				"Invalid shape for q segment_ids: "
				f"{segment_ids.q.shape}. Expected: {(q_seq_len,)}"
			)
		if segment_ids.kv.shape != (kv_seq_len,):
			raise ValueError(
				"Invalid shape for kv segment_ids: "
				f"{segment_ids.kv.shape}. Expected: {(kv_seq_len,)}"
			)

	q_layout = block_sizes.q_layout

	def q_index_map(h, i, j, data_next_ref, block_mask_ref, mask_next_ref=None):
		del j, data_next_ref, mask_next_ref, block_mask_ref
		return from_head_minor((h, i, 0), q_layout)

	def out_index_map(h, i, j, data_next_ref, block_mask_ref, mask_next_ref=None):
		del j, data_next_ref, mask_next_ref, block_mask_ref
		return h, i, 0

	k_layout = block_sizes.k_layout

	def k_index_map(h, i, j, data_next_ref, block_mask_ref, mask_next_ref=None):
		next_j, *_ = _next_nonzero(h, i, j, data_next_ref, block_mask_ref, mask_next_ref)
		prefix = () if is_mqa else (_div(h, q_heads_per_kv_head),)
		return from_head_minor((*prefix, next_j, 0), k_layout)

	v_layout = block_sizes.v_layout

	def v_index_map(h, i, j, data_next_ref, block_mask_ref, mask_next_ref=None):
		next_j, *_ = _next_nonzero(h, i, j, data_next_ref, block_mask_ref, mask_next_ref)
		prefix = () if is_mqa else (_div(h, q_heads_per_kv_head),)
		return from_head_minor((*prefix, next_j, 0), v_layout)

	def mask_index_map(h, i, j, data_next_ref, block_mask_ref, mask_next_ref=None):
		_, next_m, *_ = _next_nonzero(h, i, j, data_next_ref, block_mask_ref, mask_next_ref)
		return next_m, 0, 0

	def q_segment_ids_index_map(h, i, j, *_):
		del h, j  # Unused.
		return i, 0

	def kv_segment_ids_index_map(
		h, i, j, data_next_ref, block_mask_ref, mask_next_ref=None
	):
		next_j, *_ = _next_nonzero(h, i, j, data_next_ref, block_mask_ref, mask_next_ref)
		return 0, next_j

	# Convert the logical shape from head-minor to sequence-minor.
	in_specs = [
		pl.BlockSpec(
			from_head_minor((None, bq, head_dim), q_layout),
			q_index_map,
		),
		pl.BlockSpec(
			from_head_minor((bkv, head_dim) if is_mqa else (None, bkv, head_dim), k_layout),
			k_index_map,
		),
		pl.BlockSpec(
			from_head_minor((bkv, head_dim) if is_mqa else (None, bkv, head_dim), v_layout),
			v_index_map,
		),
	]
	if segment_ids is not None:
		in_specs += [
			pl.BlockSpec(
				(bq, NUM_LANES),
				q_segment_ids_index_map,
			),
			pl.BlockSpec(
				(NUM_SUBLANES, bkv),
				kv_segment_ids_index_map,
			),
		]
		q_segment_ids = jax.lax.broadcast_in_dim(
			segment_ids.q, (q_seq_len, NUM_LANES), (0,)
		)
		kv_segment_ids = jax.lax.broadcast_in_dim(
			segment_ids.kv, (NUM_SUBLANES, kv_seq_len), (1,)
		)
	else:
		in_specs += [None, None]
		q_segment_ids = kv_segment_ids = None

	if fwd_mask_info.partial_mask_blocks is not None:
		in_specs.append(
			pl.BlockSpec(
				(None, bq, bkv),
				mask_index_map,
			)
		)
	else:
		in_specs.append(None)

	assert fwd_mask_info.partial_mask_blocks is None or fwd_mask_info.q_sequence is None

	if fwd_mask_info.q_sequence is not None:
		q_sequence = jax.lax.broadcast_in_dim(
			fwd_mask_info.q_sequence, (q_seq_len, NUM_LANES), (0,)
		)
		in_specs.append(pl.BlockSpec((bq, NUM_LANES), q_segment_ids_index_map))
	else:
		q_sequence = None
		in_specs.append(None)

	num_scalar_prefetch = 3

	out_shapes = [
		jax.ShapeDtypeStruct((bq, NUM_LANES), jnp.float32),  # m_scratch
		jax.ShapeDtypeStruct((bq, NUM_LANES), jnp.float32),  # l_scratch
		jax.ShapeDtypeStruct((bq, head_dim), jnp.float32),  # o_scratch
		jax.ShapeDtypeStruct((num_q_heads, q_seq_len, head_dim), q.dtype),
	]
	out_specs = [
		pl.BlockSpec((bq, NUM_LANES), lambda h, i, j, *_: (0, 0)),
		pl.BlockSpec((bq, NUM_LANES), lambda h, i, j, *_: (0, 0)),
		pl.BlockSpec((bq, head_dim), lambda h, i, j, *_: (0, 0)),
		pl.BlockSpec((None, bq, head_dim), out_index_map),
	]
	if save_residuals:
		out_shapes += [
			jax.ShapeDtypeStruct(
				(num_q_heads, q_seq_len, NUM_LANES), jnp.float32
			),  # logsumexp
		]

		def logsumexp_index_map(h, i, *_):
			return h, i, 0

		out_specs += [
			pl.BlockSpec((None, bq, NUM_LANES), logsumexp_index_map),
		]
	else:
		out_shapes += [None]
		out_specs += [None]

	mosaic_params = dict(
		dimension_semantics=("parallel", "arbitrary", "arbitrary"),
		flags={"XLA_TPU_FORCE_LP_LLO_SCHEDULER": True},
	)

	kernel_name = get_kernel_name(
		dataclasses.asdict(block_sizes),
		is_mqa=is_mqa,
		save_residuals=save_residuals,
		is_segmented=segment_ids is not None,
		phase="fwd",
	)

	if fwd_mask_info.data_next is not None:
		grid_width = fwd_mask_info.data_next.shape[-1]
	else:
		grid_width = kv_seq_len // bkv

	grid = (num_q_heads, q_seq_len // bq, grid_width)
	with jax.named_scope(kernel_name):
		all_out = pl.pallas_call(
			partial(
				flash_attention_kernel,
				mask_value=mask_value,
				grid_width=grid_width,
				bq=bq,
				bkv=bkv,
				bkv_compute=bkv_compute,
				head_dim=head_dim,
				q_layout=q_layout,
				k_layout=k_layout,
				v_layout=v_layout,
				attn_logits_soft_cap=attn_logits_soft_cap,
				mask_function=mask_function,
			),
			grid_spec=pltpu.PrefetchScalarGridSpec(
				num_scalar_prefetch=num_scalar_prefetch,
				in_specs=in_specs,
				out_specs=out_specs,
				grid=grid,
			),
			compiler_params=dict(mosaic=mosaic_params),
			out_shape=out_shapes,
			name=kernel_name,
		)(
			fwd_mask_info.data_next,
			fwd_mask_info.block_mask,
			fwd_mask_info.mask_next,
			q if q_layout == QKVLayout.HEAD_DIM_MINOR else q.swapaxes(-1, -2),
			k if k_layout == QKVLayout.HEAD_DIM_MINOR else k.swapaxes(-1, -2),
			v if v_layout == QKVLayout.HEAD_DIM_MINOR else v.swapaxes(-1, -2),
			q_segment_ids,
			kv_segment_ids,
			fwd_mask_info.partial_mask_blocks,
			q_sequence,
		)

	(
		_,
		_,
		_,
		out,
		logsumexp,
	) = all_out

	if save_residuals:
		assert logsumexp is not None
		logsumexp = logsumexp[..., 0]

	if residual_checkpoint_name is not None:
		out = ad_checkpoint.checkpoint_name(out, name=residual_checkpoint_name)
		if logsumexp is not None:
			logsumexp = ad_checkpoint.checkpoint_name(
				logsumexp, name=residual_checkpoint_name
			)
	if save_residuals:
		return out, (logsumexp,)
	return out


@partial(jax.custom_vjp, nondiff_argnums=(7, 8, 9, 10, 11, 12, 13))
def _splash_attention_custom(
	fwd_mask_info: mask_info_lib.MaskInfo,
	dq_mask_info: mask_info_lib.MaskInfo | None,
	dkv_mask_info: mask_info_lib.MaskInfo | None,
	q: jax.Array,
	k: jax.Array,
	v: jax.Array,
	segment_ids: SegmentIds | None,
	save_residuals: bool,
	mask_value: float,
	is_mqa: bool,
	block_sizes: BlockSizes,
	residual_checkpoint_name: str | None,
	mask_function: MaskFunctionType | None,
	attn_logits_soft_cap: float | None = None,
) -> SplashCustomReturnType:
	# The forward function does not use the dq and dkv MaskInfos, it just forwards
	# them to the backward function as residuals. This is a way to communicate
	# arbitrary Arrays to the backward function. Since the three MaskInfos are
	# constants there is no overhead in passing them to the backward function as
	# residuals. When sharding computation MaskInfos are partitioned so both the
	# forward and the backward kernels need to work on the relevant slice. If we
	# recomputed the backward MaskInfos in the backward function from the numpy
	# mask then we would not work with the MaskInfo slice relevant to the current
	# device.
	del dq_mask_info, dkv_mask_info

	return _splash_attention_forward(  # pytype: disable=wrong-arg-types
		fwd_mask_info,
		q,
		k,
		v,
		segment_ids,
		mask_value=mask_value,
		is_mqa=is_mqa,
		block_sizes=block_sizes,
		save_residuals=save_residuals,
		attn_logits_soft_cap=attn_logits_soft_cap,
		residual_checkpoint_name=residual_checkpoint_name,
		mask_function=mask_function,
	)


def _splash_attention_fwd(
	fwd_mask_info: mask_info_lib.MaskInfo,
	dq_mask_info: mask_info_lib.MaskInfo | None,
	dkv_mask_info: mask_info_lib.MaskInfo | None,
	q: jax.Array,
	k: jax.Array,
	v: jax.Array,
	segment_ids: SegmentIds | None,
	save_residuals: bool,
	mask_value: float,
	is_mqa: bool,
	block_sizes: BlockSizes,
	residual_checkpoint_name: str | None,
	mask_function: MaskFunctionType | None,
	attn_logits_soft_cap: float | None = None,
) -> tuple[
	tuple[jax.Array],
	SplashResidualsType,
]:
	if save_residuals:
		raise NotImplementedError("Higher-order AD not supported")

	out, (logsumexp,) = _splash_attention_forward(
		fwd_mask_info,
		q,
		k,
		v,
		segment_ids,
		mask_value=mask_value,
		is_mqa=is_mqa,
		block_sizes=block_sizes,
		save_residuals=True,
		attn_logits_soft_cap=attn_logits_soft_cap,
		residual_checkpoint_name=residual_checkpoint_name,
		mask_function=mask_function,
	)
	return out, (
		q,
		k,
		v,
		segment_ids,
		out,
		logsumexp,
		dq_mask_info,
		dkv_mask_info,
	)


def _flash_attention_dq_kernel(
	# Prefetched inputs
	data_next_ref,
	block_mask_ref,
	mask_next_ref,
	# Inputs
	q_ref,
	k_ref,
	v_ref,
	q_segment_ids_ref,
	kv_segment_ids_ref,
	logsumexp_ref,
	do_ref,
	di_ref,
	mask_ref,
	q_sequence_ref,
	# Outputs
	dq_scratch_ref,
	dq_ref,
	*,
	mask_value: float,
	grid_width: int,
	bq: int,
	bkv: int,
	attn_logits_soft_cap: float | None = None,
	q_layout: QKVLayout,
	k_layout: QKVLayout,
	v_layout: QKVLayout,
	mask_function: MaskFunctionType | None,
):
	float32 = jnp.float32
	HEAD_DIM_MINOR = QKVLayout.HEAD_DIM_MINOR

	h, i, j = pl.program_id(0), pl.program_id(1), pl.program_id(2)

	@pl.when(j == 0)
	def init():
		dq_scratch_ref[...] = jnp.zeros_like(dq_scratch_ref)

	_, _, should_run, should_not_mask = _next_nonzero(
		h, i, j, data_next_ref, block_mask_ref, mask_next_ref
	)

	@pl.when(should_run)
	def run():
		q = q_ref[...] if q_layout == HEAD_DIM_MINOR else q_ref[...].T
		# We keep k and v possibly transposed, since they are RHS of dots.
		k = k_ref[...]
		v = v_ref[...]
		logsumexp = jnp.expand_dims(logsumexp_ref[0], -1)
		do = do_ref[...]
		di = jnp.expand_dims(di_ref[0], -1)

		qk_dims = NT_DIM_NUMBERS if k_layout == HEAD_DIM_MINOR else NN_DIM_NUMBERS
		qk_uncapped = lax.dot_general(q, k, qk_dims, preferred_element_type=float32)

		qk = _apply_mask_and_soft_cap(
			qk_uncapped,
			mask_value,
			should_not_mask,
			mask_ref,
			q_sequence_ref,
			q_segment_ids_ref,
			kv_segment_ids_ref,
			attn_logits_soft_cap=attn_logits_soft_cap,
			k_slice=pl.ds(0, bkv),
			k_offset=j * bkv,
			bq=bq,
			mask_function=mask_function,
		)
		p = jnp.exp(qk - logsumexp)
		dp_dims = NT_DIM_NUMBERS if v_layout == HEAD_DIM_MINOR else NN_DIM_NUMBERS
		dp = lax.dot_general(
			do.astype(v.dtype),
			v,
			dp_dims,
			preferred_element_type=jnp.float32,
		)
		ds = (dp - di) * p
		if attn_logits_soft_cap is not None:
			normalized = qk_uncapped / attn_logits_soft_cap
			d = jnp.tanh(normalized)
			g = ds * (1 - d)
			ds = g + g * d

		dq_dims = NN_DIM_NUMBERS if k_layout == HEAD_DIM_MINOR else NT_DIM_NUMBERS
		dq_scratch_ref[...] += lax.dot_general(
			ds.astype(k.dtype),
			k,
			dq_dims,
			preferred_element_type=jnp.float32,
		)

	@pl.when(j == grid_width - 1)
	def end():
		dq_ref[...] = dq_scratch_ref[...].astype(dq_ref.dtype)
		dq_scratch_ref[...] = jnp.zeros_like(dq_scratch_ref)


def _splash_attention_bwd_dq(
	q,
	k,
	v,
	segment_ids,
	logsumexp,
	do,
	di,
	*,
	bq: int,
	bkv: int,
	is_mqa: bool,
	mask_info: mask_info_lib.MaskInfo,
	mask_value: float,
	attn_logits_soft_cap: float | None,
	q_layout: QKVLayout,
	k_layout: QKVLayout,
	v_layout: QKVLayout,
	mask_function: MaskFunctionType | None,
):
	num_q_heads, q_seq_len, head_dim = q.shape
	if is_mqa:
		kv_seq_len = k.shape[0]
		num_kv_heads = 1
	else:
		kv_seq_len = k.shape[1]
		num_kv_heads = k.shape[0]

	if bq > q_seq_len:
		raise ValueError(f"{bq=} should not be greater than {q_seq_len=}")
	if bkv > kv_seq_len:
		raise ValueError(f"{bkv=} should not be greater than {kv_seq_len=}")

	if not is_mqa and num_q_heads % num_kv_heads != 0:
		raise ValueError(
			f"In MHA, expected number of 'key' heads ({num_kv_heads}) to be a"
			f" multiple of the number of 'query' heads ({num_q_heads})"
		)

	if k.shape != v.shape:
		raise ValueError(
			f"Expected 'key' {k.shape} and 'value' {v.shape} to have the same" " shape."
		)

	if bkv % NUM_LANES:
		raise ValueError(f"{bkv=} must be a multiple of {NUM_LANES}.")

	q_heads_per_kv_head = num_q_heads // num_kv_heads

	if mask_info.data_next is not None:
		grid_width = mask_info.data_next.shape[-1]
	else:
		grid_width = kv_seq_len // bkv

	grid = (num_q_heads, q_seq_len // bq, grid_width)

	def o_index_map(h, i, *_):
		return h, i, 0

	o_spec = pl.BlockSpec((None, bq, head_dim), o_index_map)

	def q_index_map(h, i, *_):
		return from_head_minor((h, i, 0), q_layout)

	q_spec = pl.BlockSpec(from_head_minor((None, bq, head_dim), q_layout), q_index_map)

	def k_index_map(h, i, j, data_next_ref, block_mask_ref, mask_next_ref, *_):
		next_j, *_ = _next_nonzero(h, i, j, data_next_ref, block_mask_ref, mask_next_ref)
		prefix = () if is_mqa else (_div(h, q_heads_per_kv_head),)
		return from_head_minor((*prefix, next_j, 0), k_layout)

	k_spec = pl.BlockSpec(
		from_head_minor((bkv, head_dim) if is_mqa else (None, bkv, head_dim), k_layout),
		k_index_map,
	)

	def v_index_map(h, i, j, data_next_ref, block_mask_ref, mask_next_ref, *_):
		next_j, *_ = _next_nonzero(h, i, j, data_next_ref, block_mask_ref, mask_next_ref)
		prefix = () if is_mqa else (_div(h, q_heads_per_kv_head),)
		return from_head_minor((*prefix, next_j, 0), v_layout)

	v_spec = pl.BlockSpec(
		from_head_minor((bkv, head_dim) if is_mqa else (None, bkv, head_dim), v_layout),
		v_index_map,
	)

	def mask_index_map(h, i, j, data_next_ref, block_mask_ref, mask_next_ref, *_):
		_, next_m, *_ = _next_nonzero(h, i, j, data_next_ref, block_mask_ref, mask_next_ref)
		return next_m, 0, 0

	mask_spec = pl.BlockSpec((None, bq, bkv), mask_index_map)

	def q_segment_ids_index_map(h, i, j, *_):
		del h, j  # Unused.
		return i, 0

	if segment_ids is not None:

		def kv_segment_ids_index_map(
			h, i, j, data_next_ref, block_mask_ref, mask_next_ref, *_
		):
			next_j, *_ = _next_nonzero(h, i, j, data_next_ref, block_mask_ref, mask_next_ref)
			return 0, next_j

		q_segment_spec = pl.BlockSpec((bq, NUM_LANES), q_segment_ids_index_map)
		kv_segment_spec = pl.BlockSpec((NUM_SUBLANES, bkv), kv_segment_ids_index_map)
		q_segment_ids = jax.lax.broadcast_in_dim(
			segment_ids.q, (q_seq_len, NUM_LANES), (0,)
		)
		kv_segment_ids = jax.lax.broadcast_in_dim(
			segment_ids.kv, (NUM_SUBLANES, kv_seq_len), (1,)
		)
	else:
		q_segment_spec = kv_segment_spec = None
		q_segment_ids = kv_segment_ids = None

	do_spec = dq_spec = o_spec

	def logsumexp_index_map(h, i, *_):
		return h, 0, i

	logsumexp = jnp.expand_dims(logsumexp, axis=-2)
	logsumexp_spec = pl.BlockSpec((None, 1, bq), logsumexp_index_map)
	assert logsumexp.ndim == len(logsumexp_spec.block_shape)

	di = jnp.expand_dims(di, axis=-2)
	di_spec = pl.BlockSpec((None, 1, bq), logsumexp_index_map)
	assert di.ndim == len(di_spec.block_shape)

	in_specs = [
		q_spec,
		k_spec,
		v_spec,
		q_segment_spec,
		kv_segment_spec,
		logsumexp_spec,
		do_spec,
		di_spec,
	]
	if mask_info.partial_mask_blocks is not None:
		in_specs.append(mask_spec)
	else:
		in_specs.append(None)

	assert mask_info.partial_mask_blocks is None or mask_info.q_sequence is None

	if mask_info.q_sequence is not None:
		q_sequence = jax.lax.broadcast_in_dim(
			mask_info.q_sequence, (q_seq_len, NUM_LANES), (0,)
		)
		in_specs.append(pl.BlockSpec(q_segment_ids_index_map, (bq, NUM_LANES)))
	else:
		q_sequence = None
		in_specs.append(None)

	out_shapes = [
		jax.ShapeDtypeStruct((bq, head_dim), jnp.float32),
		jax.ShapeDtypeStruct(q.shape, q.dtype),
	]
	out_specs = [
		pl.BlockSpec((bq, head_dim), lambda *_: (0, 0)),
		dq_spec,
	]

	kernel = functools.partial(
		_flash_attention_dq_kernel,
		grid_width=grid_width,
		mask_value=mask_value,
		bq=bq,
		bkv=bkv,
		attn_logits_soft_cap=attn_logits_soft_cap,
		q_layout=q_layout,
		k_layout=k_layout,
		v_layout=v_layout,
		mask_function=mask_function,
	)
	num_scalar_prefetch = 3

	mosaic_params = dict(
		dimension_semantics=("arbitrary", "arbitrary", "arbitrary"),
		flags={"XLA_TPU_FORCE_LP_LLO_SCHEDULER": True},
	)

	kernel_name = get_kernel_name(
		dict(
			block_q_dq=bq,
			block_kv_dq=bkv,
			q_layout=q_layout,
			k_layout=k_layout,
			v_layout=v_layout,
		),
		is_mqa=is_mqa,
		save_residuals=False,
		is_segmented=segment_ids is not None,
		phase="dq",
	)
	with jax.named_scope(kernel_name):
		_, dq = pl.pallas_call(
			kernel,
			grid_spec=pltpu.PrefetchScalarGridSpec(
				num_scalar_prefetch=num_scalar_prefetch,
				in_specs=in_specs,
				out_specs=out_specs,
				grid=grid,
			),
			out_shape=out_shapes,
			compiler_params=dict(mosaic=mosaic_params),
			name=kernel_name,
		)(
			mask_info.data_next,
			mask_info.block_mask,
			mask_info.mask_next,
			q if q_layout == QKVLayout.HEAD_DIM_MINOR else q.swapaxes(-1, -2),
			k if k_layout == QKVLayout.HEAD_DIM_MINOR else k.swapaxes(-1, -2),
			v if v_layout == QKVLayout.HEAD_DIM_MINOR else v.swapaxes(-1, -2),
			q_segment_ids,
			kv_segment_ids,
			logsumexp,
			do,
			di,
			mask_info.partial_mask_blocks,
			q_sequence,
		)
	return dq


def _flash_attention_dkv_kernel(
	# Prefetched inputs
	data_next_ref,
	block_mask_ref,
	mask_next_ref,
	# Inputs
	q_ref,
	k_ref,
	v_ref,
	q_segment_ids_ref,
	kv_segment_ids_ref,
	logsumexp_ref,
	do_ref,
	di_ref,
	mask_ref,
	q_sequence_ref,
	# Outputs
	dq_scratch_ref,
	dk_scratch_ref,
	dv_scratch_ref,
	dq_ref,
	dk_ref,
	dv_ref,
	*,
	num_q_heads: int,
	num_kv_heads: int,
	mask_value: float,
	grid_width: int,
	bq: int,
	bkv_compute: int,
	is_mqa: bool,
	attn_logits_soft_cap: float | None,
	q_layout: QKVLayout,
	k_layout: QKVLayout,
	v_layout: QKVLayout,
	bkv: int,
	mask_function: MaskFunctionType | None,
):
	HEAD_DIM_MINOR = QKVLayout.HEAD_DIM_MINOR
	kv_index, q_head_index, q_index = (
		pl.program_id(0),
		pl.program_id(1),
		pl.program_id(2),
	)
	should_initialize = q_index == 0

	q_heads_per_kv_heads = None
	q_head_index_per_kv_head = None

	# Consider this situation:
	# Q_heads:   0, 1, 2, 3, 4, 5, 6, 7
	# KV_heads:  0,    1,    2,    3
	# The gradient scratch buffers should be initialized for Q_heads 0, 2, 4, 6
	# (first Q_heads to 'see' a new KV_head).
	# The gradient output buffers should be written for Q_heads 1, 3, 5, 7 (last
	# Q_heads to 'see' the current KV_head).

	# We can use the same logic for both MQA and GA (grouped attention).
	# But for MQA there is no need for the rem instruction, so we skip it.
	if is_mqa:
		should_initialize = jnp.logical_and(should_initialize, q_head_index == 0)
	elif num_kv_heads < num_q_heads:
		q_heads_per_kv_heads = num_q_heads // num_kv_heads
		q_head_index_per_kv_head = lax.rem(q_head_index, q_heads_per_kv_heads)
		should_initialize = jnp.logical_and(
			should_initialize, q_head_index_per_kv_head == 0
		)

	@pl.when(should_initialize)
	def init():
		dk_scratch_ref[...] = jnp.zeros_like(dk_scratch_ref)
		dv_scratch_ref[...] = jnp.zeros_like(dv_scratch_ref)

	_, _, should_run, should_not_mask = _next_nonzero(
		q_head_index,
		q_index,
		kv_index,
		data_next_ref,
		block_mask_ref,
		mask_next_ref,
		next_i=True,
	)

	def body(i, _):
		slice_k = pl.ds(i * bkv_compute, bkv_compute)
		q = q_ref[...]  # We keep q potentially transposed, since it's always RHS

		def _load_kv(ref, layout):
			if layout == HEAD_DIM_MINOR:
				return pl.load(ref, (slice_k, slice(None)))
			return pl.load(ref, (slice(None), slice_k)).T

		k = _load_kv(k_ref, k_layout)
		v = _load_kv(v_ref, v_layout)
		logsumexp = pl.load(logsumexp_ref, (pl.ds(1), slice(None)))
		do = do_ref[...]
		di = pl.load(di_ref, (pl.ds(1), slice(None)))

		qk_dims = NT_DIM_NUMBERS if q_layout == HEAD_DIM_MINOR else NN_DIM_NUMBERS
		qk_uncapped = lax.dot_general(k, q, qk_dims, preferred_element_type=jnp.float32)

		qk = _apply_mask_and_soft_cap(
			qk_uncapped,
			mask_value,
			should_not_mask,
			mask_ref,
			q_sequence_ref,
			q_segment_ids_ref,
			kv_segment_ids_ref,
			attn_logits_soft_cap=attn_logits_soft_cap,
			k_slice=slice_k,
			k_offset=kv_index * bkv + i * bkv_compute,
			bq=bq,
			k_in_lanes=False,
			mask_function=mask_function,
		)
		p = jnp.exp(qk - logsumexp)
		dv = lax.dot(p.astype(do.dtype), do, preferred_element_type=jnp.float32)
		dv = dv.astype(dv_scratch_ref.dtype) + pl.load(
			dv_scratch_ref, (slice_k, slice(None))
		)
		pl.store(dv_scratch_ref, (slice_k, slice(None)), dv)

		dp = lax.dot_general(
			v,
			do,
			NT_DIM_NUMBERS,
			preferred_element_type=jnp.float32,
		)
		ds = (dp - di) * p
		if attn_logits_soft_cap is not None:
			normalized = qk_uncapped / attn_logits_soft_cap
			d = jnp.tanh(normalized)
			g = ds * (1 - d)
			ds = g + g * d
		dk_dims = NN_DIM_NUMBERS if q_layout == HEAD_DIM_MINOR else NT_DIM_NUMBERS
		dk = lax.dot_general(
			ds.astype(do.dtype), q, dk_dims, preferred_element_type=jnp.float32
		)
		dk = dk.astype(dk_scratch_ref.dtype) + pl.load(
			dk_scratch_ref, (slice_k, slice(None))
		)
		pl.store(dk_scratch_ref, (slice_k, slice(None)), dk)
		if dq_scratch_ref is not None or dq_ref is not None:
			dq = lax.dot_general(
				ds.T.astype(k.dtype),
				k,
				NN_DIM_NUMBERS,
				preferred_element_type=jnp.float32,
			)
			if dq_scratch_ref is not None:
				# Compute block size != memory block size
				dq_scratch_ref[...] += dq
			else:
				# Compute block size == memory block size
				assert dq_ref is not None
				dq_ref[...] = dq.astype(dq_ref.dtype)

	if dq_scratch_ref is not None:
		dq_scratch_ref[...] = jnp.zeros_like(dq_scratch_ref)
	elif dq_scratch_ref is None and dq_ref is not None:
		dq_ref[...] = jnp.zeros_like(dq_ref)

	@pl.when(should_run)
	def run():
		num_iters = k_ref.shape[0 if k_layout is HEAD_DIM_MINOR else 1] // bkv_compute
		lax.fori_loop(0, num_iters, body, None, unroll=True)

	if dq_scratch_ref is not None:
		assert dq_ref is not None
		dq_ref[...] = dq_scratch_ref[...].astype(dq_ref.dtype)

	should_write = q_index == grid_width - 1
	if is_mqa:
		should_write = jnp.logical_and(should_write, q_head_index == num_q_heads - 1)
	elif num_kv_heads < num_q_heads:
		should_write = jnp.logical_and(
			should_write, q_head_index_per_kv_head == q_heads_per_kv_heads - 1
		)

	@pl.when(should_write)
	def end():
		dk_ref[...] = dk_scratch_ref[...].astype(dk_ref.dtype)
		dv_ref[...] = dv_scratch_ref[...].astype(dv_ref.dtype)
		if dq_scratch_ref is not None:
			dq_scratch_ref[...] = jnp.zeros_like(dq_scratch_ref)

		dk_scratch_ref[...] = jnp.zeros_like(dk_scratch_ref)
		dv_scratch_ref[...] = jnp.zeros_like(dv_scratch_ref)


def _splash_attention_bwd_dkv(
	q,
	k,
	v,
	segment_ids,
	logsumexp,
	do,
	di,
	*,
	bq: int,
	bkv: int,
	bkv_compute: int,
	is_mqa: bool,
	mask_info: mask_info_lib.MaskInfo,
	mask_value: float,
	attn_logits_soft_cap: float | None,
	use_fused_bwd_kernel: bool,
	q_layout: QKVLayout,
	k_layout: QKVLayout,
	v_layout: QKVLayout,
	mask_function: MaskFunctionType | None,
):
	num_q_heads, q_seq_len, head_dim = q.shape
	if is_mqa:
		num_kv_heads, kv_seq_len = 1, k.shape[0]
	else:
		num_kv_heads, kv_seq_len, _ = k.shape

	if bq > q_seq_len:
		raise ValueError(f"{bq=} should not be greater than {q_seq_len=}")
	if bkv > kv_seq_len:
		raise ValueError(f"{bkv=} should not be greater than {kv_seq_len=}")
	if bkv_compute > bkv:
		raise ValueError(f"{bkv_compute=} should not be greater than {bkv=}")
	if bkv % bkv_compute:
		raise ValueError(f"{bkv=} should be a multiple of {bkv_compute=}")

	if not is_mqa and num_q_heads % num_kv_heads != 0:
		raise ValueError(
			f"In MHA, expected number of 'key' heads ({num_kv_heads}) to be a"
			f" multiple of the number of 'query' heads ({num_q_heads})"
		)

	if k.shape != v.shape:
		raise ValueError(
			f"Expected 'key' {k.shape} and 'value' {v.shape} to have the same" " shape."
		)

	q_heads_per_kv_head = num_q_heads // num_kv_heads

	if mask_info.data_next is not None:
		grid_width = mask_info.data_next.shape[-2]
	else:
		grid_width = q_seq_len // bq

	grid = (
		kv_seq_len // bkv,
		num_q_heads,
		grid_width,
	)

	def o_index_map(
		kv_index,
		head_index,
		q_index,
		data_next_ref,
		block_mask_ref,
		mask_next_ref=None,
	):
		next_i, *_ = _next_nonzero(
			head_index,
			q_index,
			kv_index,
			data_next_ref,
			block_mask_ref,
			mask_next_ref,
			next_i=True,
		)
		return head_index, next_i, 0

	o_spec = pl.BlockSpec((None, bq, head_dim), o_index_map)

	def q_index_map(
		kv_index,
		head_index,
		q_index,
		data_next_ref,
		block_mask_ref,
		mask_next_ref=None,
	):
		next_i, *_ = _next_nonzero(
			head_index,
			q_index,
			kv_index,
			data_next_ref,
			block_mask_ref,
			mask_next_ref,
			next_i=True,
		)
		return from_head_minor((head_index, next_i, 0), q_layout)

	q_spec = pl.BlockSpec(from_head_minor((None, bq, head_dim), q_layout), q_index_map)

	def k_index_map(kv_index, head_index, *_):
		prefix = () if is_mqa else (_div(head_index, q_heads_per_kv_head),)
		return from_head_minor((*prefix, kv_index, 0), k_layout)

	k_spec = pl.BlockSpec(
		from_head_minor(
			(bkv, head_dim) if is_mqa else (None, bkv, head_dim),
			k_layout,
		),
		k_index_map,
	)

	def v_index_map(kv_index, head_index, *_):
		prefix = () if is_mqa else (_div(head_index, q_heads_per_kv_head),)
		return from_head_minor((*prefix, kv_index, 0), v_layout)

	v_spec = pl.BlockSpec(
		from_head_minor(
			(bkv, head_dim) if is_mqa else (None, bkv, head_dim),
			v_layout,
		),
		v_index_map,
	)

	if use_fused_bwd_kernel:

		def dq_index_map(kv_index, head_index, q_index, *_):
			return (kv_index, head_index, q_index, 0)

		dq_spec = pl.BlockSpec((None, None, bq, head_dim), dq_index_map)
		dq_shape = jax.ShapeDtypeStruct((kv_seq_len // bkv, *q.shape), q.dtype)
		if bkv == bkv_compute:
			dq_scratch_spec = dq_scratch_shape = None
		else:
			dq_scratch_spec = pl.BlockSpec((bq, head_dim), lambda *_: (0, 0))
			dq_scratch_shape = jax.ShapeDtypeStruct((bq, head_dim), jnp.float32)
	else:
		dq_spec = dq_shape = dq_scratch_spec = dq_scratch_shape = None

	def dkv_index_map(kv_index, head_index, *_):
		prefix = () if is_mqa else (_div(head_index, q_heads_per_kv_head),)
		return (*prefix, kv_index, 0)

	dk_spec = dv_spec = pl.BlockSpec(
		(bkv, head_dim) if is_mqa else (None, bkv, head_dim),
		dkv_index_map,
	)

	def mask_index_map(
		kv_index,
		head_index,
		q_index,
		data_next_ref,
		block_mask_ref,
		mask_next_ref,
	):
		_, next_m, *_ = _next_nonzero(
			head_index,
			q_index,
			kv_index,
			data_next_ref,
			block_mask_ref,
			mask_next_ref,
			next_i=True,
		)
		return next_m, 0, 0

	mask_spec = pl.BlockSpec((None, bkv, bq), mask_index_map)

	def q_segment_ids_index_map(
		kv_index,
		head_index,
		q_index,
		data_next_ref,
		block_mask_ref,
		mask_next_ref=None,
	):
		next_i, *_ = _next_nonzero(
			head_index,
			q_index,
			kv_index,
			data_next_ref,
			block_mask_ref,
			mask_next_ref,
			next_i=True,
		)
		return 0, next_i

	if segment_ids is not None:

		def kv_segment_ids_index_map(kv_index, *_):
			return kv_index, 0

		q_segment_spec = pl.BlockSpec((NUM_SUBLANES, bq), q_segment_ids_index_map)
		kv_segment_spec = pl.BlockSpec((bkv, NUM_LANES), kv_segment_ids_index_map)
		q_segment_ids = jax.lax.broadcast_in_dim(
			segment_ids.q, (NUM_SUBLANES, q_seq_len), (1,)
		)
		kv_segment_ids = jax.lax.broadcast_in_dim(
			segment_ids.kv, (kv_seq_len, NUM_LANES), (0,)
		)
	else:
		q_segment_spec = kv_segment_spec = None
		q_segment_ids = kv_segment_ids = None

	do_spec = o_spec

	def logsumexp_index_map(
		kv_index,
		head_index,
		q_index,
		data_next_ref,
		block_mask_ref,
		mask_next_ref=None,
	):
		next_i, *_ = _next_nonzero(
			head_index,
			q_index,
			kv_index,
			data_next_ref,
			block_mask_ref,
			mask_next_ref,
			next_i=True,
		)
		return head_index, 0, next_i

	assert logsumexp.shape == di.shape == (num_q_heads, q_seq_len)
	logsumexp_shape = (num_q_heads, NUM_SUBLANES, q_seq_len)
	logsumexp = jnp.broadcast_to(jnp.expand_dims(logsumexp, -2), logsumexp_shape)
	logsumexp_spec = pl.BlockSpec(
		(None, NUM_SUBLANES, bq),
	)
	assert logsumexp.ndim == len(logsumexp_spec.block_shape)
	di = jnp.broadcast_to(jnp.expand_dims(di, -2), logsumexp_shape)
	di_spec = pl.BlockSpec((None, NUM_SUBLANES, bq), logsumexp_index_map)
	assert di.ndim == len(di_spec.block_shape)

	in_specs = [
		q_spec,
		k_spec,
		v_spec,
		q_segment_spec,
		kv_segment_spec,
		logsumexp_spec,
		do_spec,
		di_spec,
	]
	if mask_info.partial_mask_blocks is not None:
		in_specs.append(mask_spec)
	else:
		in_specs.append(None)

	if mask_info.q_sequence is not None:
		in_specs.append(pl.BlockSpec((NUM_SUBLANES, bq), q_segment_ids_index_map))
		q_sequence = jax.lax.broadcast_in_dim(
			mask_info.q_sequence, (NUM_SUBLANES, q_seq_len), (1,)
		)
	else:
		q_sequence = None
		in_specs.append(None)

	out_shapes = [
		dq_scratch_shape,
		jax.ShapeDtypeStruct((bkv, head_dim), jnp.float32),
		jax.ShapeDtypeStruct((bkv, head_dim), jnp.float32),
		dq_shape,
		jax.ShapeDtypeStruct(k.shape, k.dtype),
		jax.ShapeDtypeStruct(v.shape, v.dtype),
	]
	out_specs = [
		dq_scratch_spec,
		pl.BlockSpec((bkv, head_dim), lambda *_: (0, 0)),
		pl.BlockSpec((bkv, head_dim), lambda *_: (0, 0)),
		dq_spec,
		dk_spec,
		dv_spec,
	]

	kernel = functools.partial(
		_flash_attention_dkv_kernel,
		mask_value=mask_value,
		num_q_heads=num_q_heads,
		num_kv_heads=num_kv_heads,
		is_mqa=is_mqa,
		grid_width=grid_width,
		bq=bq,
		bkv_compute=bkv_compute,
		attn_logits_soft_cap=attn_logits_soft_cap,
		q_layout=q_layout,
		k_layout=k_layout,
		v_layout=v_layout,
		bkv=bkv,
		mask_function=mask_function,
	)
	num_scalar_prefetch = 3

	# We set all dimensions to arbitrary because:
	# 1) for kv_seq_len, the splash attention prefetch schedule assumes no
	#    megacore
	# 2) for heads, we are reducing over heads
	# 3) for q_seq_len, we are reducing over it to compute dkv
	mosaic_params = dict(
		dimension_semantics=("arbitrary", "arbitrary", "arbitrary"),
		flags={"XLA_TPU_FORCE_LP_LLO_SCHEDULER": True},
	)

	kernel_name = get_kernel_name(
		dict(
			block_q_dkv=bq,
			block_kv_dkv=bkv,
			block_kv_dkv_compute=bkv_compute,
			q_layout=q_layout,
			k_layout=k_layout,
			v_layout=v_layout,
		),
		is_mqa=is_mqa,
		save_residuals=False,
		is_segmented=segment_ids is not None,
		phase="dkv",
	)
	with jax.named_scope(kernel_name):
		_, _, _, dq_unreduced, dk, dv = pl.pallas_call(
			kernel,
			grid_spec=pltpu.PrefetchScalarGridSpec(
				num_scalar_prefetch=num_scalar_prefetch,
				in_specs=in_specs,
				out_specs=out_specs,
				grid=grid,
			),
			out_shape=out_shapes,
			compiler_params=dict(mosaic=mosaic_params),
			name=kernel_name,
		)(
			mask_info.data_next,
			mask_info.block_mask,
			mask_info.mask_next,
			q if q_layout == QKVLayout.HEAD_DIM_MINOR else q.swapaxes(-1, -2),
			k if k_layout == QKVLayout.HEAD_DIM_MINOR else k.swapaxes(-1, -2),
			v if v_layout == QKVLayout.HEAD_DIM_MINOR else v.swapaxes(-1, -2),
			q_segment_ids,
			kv_segment_ids,
			logsumexp,
			do,
			di,
			mask_info.partial_mask_blocks,
			q_sequence,
		)
	if use_fused_bwd_kernel:
		assert dq_unreduced is not None
		dq = dq_unreduced.sum(axis=0)
	else:
		assert dq_unreduced is None
		dq = None
	return dq, dk, dv


def _splash_attention_bwd(
	save_residuals: bool,
	mask_value: float,
	is_mqa: bool,
	block_sizes: BlockSizes,
	residual_checkpoint_name: str | None,
	mask_function: MaskFunctionType | None,
	attn_logits_soft_cap: float | None,
	res: SplashResidualsType,
	do: jax.Array,
) -> tuple[
	mask_info_lib.MaskInfo | None,  # fwd_mask_info
	mask_info_lib.MaskInfo | None,  # dq_mask_info
	mask_info_lib.MaskInfo | None,  # dvk_mask_info
	jax.Array,  # q
	jax.Array,  # k
	jax.Array,  # v
	SegmentIds | None,  # segmend_ids
]:
	del save_residuals, residual_checkpoint_name
	if not block_sizes.has_backward_blocks:
		raise ValueError("Need to specify backward blocks.")
	bq_dq, bkv_dq = block_sizes.block_q_dq, block_sizes.block_kv_dq
	bq_dkv, bkv_dkv_memory, bkv_dkv_compute = (
		block_sizes.block_q_dkv,
		block_sizes.block_kv_dkv,
		block_sizes.block_kv_dkv_compute,
	)
	use_fused_bwd_kernel = block_sizes.use_fused_bwd_kernel
	(
		q,
		k,
		v,
		segment_ids,
		o,
		logsumexp,
		dq_mask_info,
		dkv_mask_info,
	) = res

	# di: [num_heads, q_seq_len]
	di = jnp.einsum(
		"hsd,hsd->hs", o.astype(jnp.float32), do.astype(jnp.float32)
	)  # pytype: disable=attribute-error
	dq, dk, dv = _splash_attention_bwd_dkv(
		q,
		k,
		v,
		segment_ids,
		logsumexp,
		do,
		di,
		bq=bq_dkv,
		bkv=bkv_dkv_memory,
		bkv_compute=bkv_dkv_compute,
		is_mqa=is_mqa,
		mask_info=dkv_mask_info,
		mask_value=mask_value,
		attn_logits_soft_cap=attn_logits_soft_cap,
		use_fused_bwd_kernel=use_fused_bwd_kernel,
		q_layout=block_sizes.q_layout,
		k_layout=block_sizes.k_layout,
		v_layout=block_sizes.v_layout,
		mask_function=mask_function,
	)
	if not use_fused_bwd_kernel:
		assert dq is None
		dq = _splash_attention_bwd_dq(
			q,
			k,
			v,
			segment_ids,
			logsumexp,
			do,
			di,
			bq=bq_dq,
			bkv=bkv_dq,
			is_mqa=is_mqa,
			mask_info=dq_mask_info,
			mask_value=mask_value,
			attn_logits_soft_cap=attn_logits_soft_cap,
			q_layout=block_sizes.q_layout,
			k_layout=block_sizes.k_layout,
			v_layout=block_sizes.v_layout,
			mask_function=mask_function,
		)
	# Match the signature of the fwd function.
	assert dq is not None
	return (
		None,  # fwd_mask_info
		None,  # dq_mask_info
		None,  # dvk_mak_info
		dq,  # q
		dk,  # k
		dv,  # v
		None,  # segment_ids
	)


_splash_attention_custom.defvjp(_splash_attention_fwd, _splash_attention_bwd)


@partial(
	jax.jit,
	static_argnames=[
		"is_mqa",
		"block_sizes",
		"save_residuals",
		"mask_value",
		"attn_logits_soft_cap",
		"residual_checkpoint_name",
		"mask_function",
	],
)
def _splash_attention(
	fwd_mask_info: mask_info_lib.MaskInfo,
	dq_mask_info: mask_info_lib.MaskInfo | None,
	dkv_mask_info: mask_info_lib.MaskInfo | None,
	q: jax.Array,
	k: jax.Array,
	v: jax.Array,
	segment_ids: SegmentIds | None = None,
	*,
	is_mqa: bool,
	block_sizes: BlockSizes | None,
	save_residuals: bool,
	mask_value: float,
	attn_logits_soft_cap: float | None,
	residual_checkpoint_name: str | None,
	mask_function: MaskFunctionType | None,
) -> SplashCustomReturnType:
	return _splash_attention_custom(
		fwd_mask_info,
		dq_mask_info,
		dkv_mask_info,
		q,
		k,
		v,
		segment_ids,
		mask_value=mask_value,
		is_mqa=is_mqa,
		block_sizes=block_sizes,
		save_residuals=save_residuals,
		attn_logits_soft_cap=attn_logits_soft_cap,
		residual_checkpoint_name=residual_checkpoint_name,
		mask_function=mask_function,
	)


@jax.tree_util.register_pytree_node_class
class SplashAttentionKernel:
	def __init__(
		self,
		fwd_mask_info: mask_info_lib.MaskInfo,
		dq_mask_info: mask_info_lib.MaskInfo | None,
		dkv_mask_info: mask_info_lib.MaskInfo | None,
		**kwargs,
	):
		self.kwargs = kwargs
		self.fwd_mask_info = fwd_mask_info
		self.dq_mask_info = dq_mask_info
		self.dkv_mask_info = dkv_mask_info

	def __call__(self, *args, **kwargs) -> SplashCustomReturnType:
		return _splash_attention(
			self.fwd_mask_info,
			self.dq_mask_info,
			self.dkv_mask_info,
			*args,
			**kwargs,
			**self.kwargs,
		)

	def manual_sharding_spec(self, sharding: jax.sharding.NamedSharding):
		"""Returns a value that can be used as a shard_map partition spec for the kernel."""
		if self.fwd_mask_info.data_next is not None:
			block_mask_shape = self.fwd_mask_info.data_next.shape
			try:
				shard_shape = sharding.shard_shape(block_mask_shape)
			except ValueError as exc:
				raise ValueError(
					"The sharding must divide the mask blocks evenly between devices"
				) from exc
			if block_mask_shape[-1] != shard_shape[-1]:
				raise ValueError("Sharding the kv sequence dimension is not supported")
		spec = sharding.spec
		assert len(spec) == 2
		replicated = jax.sharding.PartitionSpec()
		# Shard q_sequence over the sequence dimension only.
		q_sequence_spec = jax.sharding.PartitionSpec(spec[1])
		mask_info_specs = mask_info_lib.MaskInfo(  # pytype: disable=wrong-arg-types
			data_next=spec if self.fwd_mask_info.data_next is not None else None,
			mask_next=spec if self.fwd_mask_info.mask_next is not None else None,
			block_mask=spec if self.fwd_mask_info.block_mask is not None else None,
			partial_mask_blocks=replicated
			if self.fwd_mask_info.partial_mask_blocks is not None
			else None,
			q_sequence=q_sequence_spec if self.fwd_mask_info.q_sequence is not None else None,
		)
		return SplashAttentionKernel(
			mask_info_specs,
			mask_info_specs if self.dq_mask_info is not None else None,
			mask_info_specs if self.dkv_mask_info is not None else None,
			**self.kwargs,
		)

	def tree_flatten(self):
		return (
			(self.fwd_mask_info, self.dq_mask_info, self.dkv_mask_info),
			self.kwargs,
		)

	@classmethod
	def tree_unflatten(cls, kwargs, values):
		fwd_mask_info, dq_mask_info, dkv_mask_info = values
		# NamedTuples are not preserved during pytree serialization.
		dq_mask_info = (
			mask_info_lib.MaskInfo(*dq_mask_info) if dq_mask_info is not None else None
		)
		dkv_mask_info = (
			mask_info_lib.MaskInfo(*dkv_mask_info) if dkv_mask_info is not None else None
		)
		return SplashAttentionKernel(
			mask_info_lib.MaskInfo(*fwd_mask_info),
			dq_mask_info,
			dkv_mask_info,
			**kwargs,
		)


def _make_splash_attention(
	mask: np.ndarray | mask_lib.MultiHeadMask,
	*,
	block_sizes: BlockSizes | None = None,
	is_mqa: bool,
	save_residuals: bool = False,
	mask_value: float = DEFAULT_MASK_VALUE,
	attn_logits_soft_cap: float | None = None,
	downcast_smem_data: bool = True,
	head_shards: int,
	q_seq_shards: int,
	residual_checkpoint_name: str | None = None,
):
	if len(mask.shape) != 3:
		raise ValueError(f"Unexpected mask shape: {mask.shape}")

	if isinstance(mask, np.ndarray):
		mask = mask_lib.MultiHeadMask([mask_lib.NumpyMask(head_mask) for head_mask in mask])

	if block_sizes is None:
		block_sizes = BlockSizes.get_default()
	fwd_mask_info, mask_function_fwd = mask_info_lib.process_mask(
		mask,
		(block_sizes.block_q, block_sizes.block_kv),
		downcast_smem_data=downcast_smem_data,
		head_shards=head_shards,
		q_seq_shards=q_seq_shards,
	)

	fwd_mask_info = tree_util.tree_map(jnp.array, fwd_mask_info)

	dq_mask_info = None
	dkv_mask_info = None
	if block_sizes.has_backward_blocks:
		if block_sizes.use_fused_bwd_kernel:
			dq_mask_info = None
		else:
			bq_dq, bkv_dq = block_sizes.block_q_dq, block_sizes.block_kv_dq
			dq_mask_info, mask_function_dq = mask_info_lib.process_mask(
				mask,
				(bq_dq, bkv_dq),
				downcast_smem_data=downcast_smem_data,
				head_shards=head_shards,
				q_seq_shards=q_seq_shards,
			)
			assert (mask_function_fwd is None) == (mask_function_dq is None)
			dq_mask_info = tree_util.tree_map(jnp.array, dq_mask_info)
		bq_dkv, bkv_dkv = block_sizes.block_q_dkv, block_sizes.block_kv_dkv
		dkv_mask_info, mask_function_dkv = mask_info_lib.process_mask_dkv(
			mask,
			(bq_dkv, bkv_dkv),
			downcast_smem_data=downcast_smem_data,
			head_shards=head_shards,
			q_seq_shards=q_seq_shards,
			shrink_grid=not block_sizes.use_fused_bwd_kernel,
		)
		assert (mask_function_fwd is None) == (mask_function_dkv is None)

		dkv_mask_info = tree_util.tree_map(jnp.array, dkv_mask_info)

	return SplashAttentionKernel(
		fwd_mask_info,
		dq_mask_info,
		dkv_mask_info,
		block_sizes=block_sizes,
		is_mqa=is_mqa,
		save_residuals=save_residuals,
		mask_value=mask_value,
		attn_logits_soft_cap=attn_logits_soft_cap,
		residual_checkpoint_name=residual_checkpoint_name,
		mask_function=mask_function_fwd,
	)


make_splash_mha = partial(_make_splash_attention, is_mqa=False)
make_splash_mqa = partial(_make_splash_attention, is_mqa=True)

make_splash_mha_single_device = partial(
	make_splash_mha, is_mqa=False, head_shards=1, q_seq_shards=1
)

make_splash_mqa_single_device = partial(
	make_splash_mha, is_mqa=True, head_shards=1, q_seq_shards=1
)



# jax.experimental.pallas.ops.gpu

"""Pallas softmax kernel."""

import functools

import jax
import jax.numpy as jnp
from jax.experimental import pallas as pl


def _vmappable_softmax_kernel(
	# inputs
	input_ref,
	# outputs
	probs_ref,
	*,
	# block information
	# It is assumed that block_row >= row_len
	block_row: int,
):
	row_len = input_ref.shape[-1]

	mask = jnp.arange(block_row) < row_len
	row = pl.load(input_ref, (pl.dslice(0, block_row),), mask=mask, other=-float("inf"))

	row_max = jnp.max(row, axis=0)
	numerator = jnp.exp((row - row_max).astype(jnp.float32))
	denominator = jnp.sum(numerator, axis=0)

	pl.store(
		probs_ref,
		(pl.dslice(0, block_row),),
		(numerator / denominator).astype(probs_ref.dtype),
		mask=mask,
	)


@functools.partial(jax.jit, static_argnames=["axis", "num_warps", "interpret", "debug"])
def softmax(
	x: jax.Array,
	*,
	axis: int = -1,
	num_warps: int = 4,
	interpret: bool = False,
	debug: bool = False,
) -> jax.Array:
	"""Computes the softmax of the input array along the specified axis.

	Args:
	  x: input array
	  axis: the axis along which to perform the computation
	  num_warps: the number of warps to use for executing the Triton kernel
	  interpret: whether to interpret the kernel using pallas
	  debug: whether to use pallas in debug mode

	Returns:
	  The result of the softmax operation over the specified axis of x.
	"""
	axis = axis if axis >= 0 else len(x.shape) + axis
	if axis != len(x.shape) - 1:
		raise NotImplementedError("reductions along non-trailing dimension unsupported")

	row_len = x.shape[-1]

	block_row = pl.next_power_of_2(row_len)
	out_shape = jax.ShapeDtypeStruct(shape=(row_len,), dtype=x.dtype)

	kernel = functools.partial(_vmappable_softmax_kernel, block_row=block_row)
	f = pl.pallas_call(
		kernel,
		compiler_params=dict(triton=dict(num_warps=num_warps, num_stages=1)),
		grid=(),
		out_shape=out_shape,
		debug=debug,
		interpret=interpret,
	)

	for _ in range(len(x.shape) - 1):
		f = jax.vmap(f)

	return f(x)



from fjformer.pallas_operations.tpu import flash_attention as flash_attention
from fjformer.pallas_operations.tpu import paged_attention as paged_attention
from fjformer.pallas_operations.tpu import ring_attention as ring_attention
from fjformer.pallas_operations.tpu import splash_attention as splash_attention



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Numerics for int8, int4, binary and other integer types."""

from typing import Any, Optional
from fjformer.bit_quantization import stochastic_rounding
from fjformer.bit_quantization import numerics
import flax.struct
from jax import lax
import jax.numpy as jnp


class IntNumerics(numerics.QNumerics, flax.struct.PyTreeNode):
	"""Numerics for int8, int4, binary, etc."""

	bits: int
	preserve_zero: bool
	preserve_max_val: bool
	clip: bool
	clip_gradient: bool
	round: bool
	noise_fn: Optional[stochastic_rounding.NoiseFn]
	dtype: Optional[Any] = None

	def get_edge_of_last_int_bucket(self):
		ret = 2.0 ** (self.bits - 1)
		if self.preserve_zero:
			# Lose one bucket.
			ret -= 0.5
		return ret

	def get_center_of_last_int_bucket(self):
		return self.get_edge_of_last_int_bucket() - 0.5

	def abs_val_mapped_to(self):
		if self.preserve_max_val:
			return self.get_center_of_last_int_bucket()
		else:
			return self.get_edge_of_last_int_bucket()

	def _get_fwd_clip_bound(self):
		# If we are not rounding, we just clip to bucket edges.
		fwd_clip_bound = self.get_edge_of_last_int_bucket()
		# If, after clip, we are rounding, we need to make sure that
		# we won't round values at the clip_bound away to the
		# non-existing bucket.
		if self.round:
			# Reducing fwd_clip_bound by any value in (0.0, 1.0) is correct.
			fwd_clip_bound -= 0.5
		return fwd_clip_bound

	def get_dtype(self):
		return self.dtype

	def fwd(self, x, context):
		"""Forward pass."""
		assert self.bits <= 22, "Too many bits, float32 has less precision."
		# Maybe noise
		if self.noise_fn:
			assert context.key is not None, (
				"noise_fn is set, requestic stochastic rounding, but RNG was not "
				"passed in Context.key"
			)
			x = (x + self.noise_fn(x.shape, context.key)).astype(x.dtype)

		if self.clip:
			fwd_clip_bound = self._get_fwd_clip_bound()
			x = jnp.clip(x, -fwd_clip_bound, fwd_clip_bound)

		# Maybe round
		if self.round:
			round_to_halves = not self.preserve_zero
			if round_to_halves:
				x = jnp.floor(x) + 0.5
			else:
				x = lax.round(x, lax.RoundingMethod.TO_NEAREST_EVEN)

		return x

	def vjp_fwd(self, x, context):
		res = (x,)
		return self.fwd(x, context), res

	def vjp_bwd(self, res, grad):
		# Gradient of the clip function.
		# For boundary values we will have full gradient.
		# When using abs(max(x)) scaling, x is always in the interior, and the
		# gradient clip is always 1. So, we can always set clip_gradient to false.
		# However, other types of scaling may result in x being outside (i.e., there
		# is clipping). In that case it may be desirable to make the gradient zero.
		ret = grad
		if self.clip_gradient:
			(x,) = res
			clip_bound = self._get_fwd_clip_bound()
			ret *= (-clip_bound <= x) * (x <= clip_bound)
		return (ret, None)



# Modified Implementation of Flash attention
from __future__ import annotations

import functools
import math
from typing import Any, Optional

import jax
import jax.numpy as jnp
import numpy as np
from jax import lax
from jax.experimental import pallas as pl

DEFAULT_MASK_VALUE = -0.7 * float(np.finfo(np.dtype("float32")).max)


def flash_attention_forward_kernel(
	q_ref,
	k_ref,
	v_ref,  # Input arrays
	b_ref: jax.Array | None,  # bias
	o_ref: Any,  # Output
	*residual_refs: Any,  # Residual outputs
	num_heads: int,
	sm_scale: float,
	block_q: int,
	block_d: int,
	block_k: int,
):
	seq_len = q_ref.shape[0]
	start_q = pl.program_id(0)
	if sm_scale is None:
		sm_scale = 1 / math.sqrt(q_ref.shape[-1])
	m_i = jnp.zeros(block_q, dtype=jnp.float32) - float("inf")
	l_i = jnp.zeros(block_q, dtype=jnp.float32)
	o = jnp.zeros((block_q, block_d), dtype=jnp.float32)

	curr_q_slice = pl.dslice(start_q * block_q, block_q)
	q = pl.load(q_ref, (curr_q_slice, pl.dslice(None)))

	def body(start_k, carry):
		o_prev, m_prev, l_prev = carry
		curr_k_slice = pl.dslice(start_k * block_k, block_k)

		k = pl.load(k_ref, (curr_k_slice, slice(None)))

		qk = pl.dot(q, k.T)
		if sm_scale != 1.0:
			qk *= sm_scale

		if b_ref is not None:
			b = pl.load(b_ref, (curr_q_slice, curr_k_slice))
			qk = jnp.add(
				b,
				qk,
			)

		m_curr = qk.max(axis=-1)
		m_next = jnp.maximum(m_prev, m_curr)
		correction = jnp.exp(m_prev - m_next)
		l_prev_corr = correction * l_prev
		s_curr = jnp.exp(qk - m_next[:, None])
		l_curr = s_curr.sum(axis=-1)
		l_next = l_prev_corr + l_curr
		l_next_rcp = 1.0 / l_next
		s_curr = s_curr * l_next_rcp[:, None]
		o_prev_corr = (l_prev_corr * l_next_rcp)[:, None] * o_prev
		v = pl.load(v_ref, (curr_k_slice, pl.dslice(block_d)))
		o_curr = pl.dot(s_curr.astype(v.dtype), v)

		o_next = o_prev_corr + o_curr
		return o_next, m_next, l_next

	upper_bound = pl.cdiv(seq_len, block_k)
	o, m_i, l_i = lax.fori_loop(0, upper_bound, body, (o, m_i, l_i))

	if residual_refs:
		l_ref, m_ref = residual_refs
		pl.store(l_ref, (curr_q_slice,), l_i)
		pl.store(m_ref, (curr_q_slice,), m_i)
	# Write output to dram.
	o = o.astype(o_ref.dtype)
	pl.store(o_ref, (curr_q_slice, pl.dslice(None)), o)


@functools.partial(jax.custom_vjp, nondiff_argnums=[4, 5, 6, 7, 8, 9, 10, 11, 12])
@functools.partial(
	jax.jit,
	static_argnames=[
		"sm_scale",
		"block_q",
		"block_k",
		"backward_pass_impl",
		"num_warps",
		"num_stages",
		"grid",
		"interpret",
		"debug",
	],
)
def flash_attention(
	query,
	key,
	value,
	bias: Optional[jnp.ndarray] = None,
	sm_scale: Optional[float] = None,
	block_q: int = 128,
	block_k: int = 128,
	backward_pass_impl: str = "triton",
	num_warps: Optional[int] = None,
	num_stages: int = 2,
	grid: Optional[tuple[int, ...]] = None,
	interpret: Optional[bool] = None,
	debug: bool = False,
):
	del backward_pass_impl

	batch_size, seq_len, num_heads, head_dim = query.shape
	if sm_scale is None:
		sm_scale = 1 / math.sqrt(head_dim)
	if interpret is None:
		interpret = (
			not (seq_len / 16).is_integer()
			or jax.lib.xla_bridge.get_backend().platform == "cpu"
		)

	block_q = min(block_q, seq_len)
	block_k = min(block_k, seq_len)
	# Heuristics.
	grid_ = grid
	if grid_ is None:
		grid_ = (pl.cdiv(seq_len, block_q), batch_size, num_heads)

	num_warps_ = num_warps
	if num_warps_ is None:
		num_warps_ = 4 if head_dim <= 64 else 8
	kernel = functools.partial(
		flash_attention_forward_kernel,
		num_heads=num_heads,
		sm_scale=sm_scale,
		block_q=block_q,
		block_k=block_k,
		block_d=head_dim,
	)

	in_specs = [
		pl.BlockSpec((None, seq_len, None, head_dim), lambda _, j, k: (j, 0, k, 0)),
		pl.BlockSpec((None, seq_len, None, head_dim), lambda _, j, k: (j, 0, k, 0)),
		pl.BlockSpec((None, seq_len, None, head_dim), lambda _, j, k: (j, 0, k, 0)),
		(
			None
			if bias is None
			else pl.BlockSpec((None, None, seq_len, seq_len), lambda _, j, k: (j, 0, 0, 0))
		),
	]
	out_shape = jax.ShapeDtypeStruct(shape=query.shape, dtype=query.dtype)
	return pl.pallas_call(
		kernel,
		grid=grid_,
		in_specs=in_specs,
		out_specs=pl.BlockSpec(
			(None, seq_len, None, head_dim),
			lambda _, j, k: (j, 0, k, 0),
		),
		compiler_params=dict(triton=dict(num_warps=num_warps_, num_stages=num_stages)),
		out_shape=out_shape,
		debug=debug,
		interpret=interpret,
		name="flash_attention_forward",
	)(query, key, value, bias)


def _flash_attention_forward(
	query,
	key,
	value,
	bias: jax.Array | None,
	sm_scale: float,
	block_q: int,
	block_k: int,
	backward_pass_impl: str,
	num_warps: int | None,
	num_stages: int,
	grid: Any,
	interpret: bool,
	debug: bool,
):
	del backward_pass_impl
	batch_size, seq_len, num_heads, head_dim = query.shape
	if sm_scale is None:
		sm_scale = 1 / math.sqrt(head_dim)
	if interpret is None:
		interpret = (
			not (seq_len / 16).is_integer()
			or jax.lib.xla_bridge.get_backend().platform == "cpu"
		)
	block_q = min(block_q, seq_len)
	block_k = min(block_k, seq_len)
	# Heuristics.
	grid_ = grid
	if grid_ is None:
		grid_ = (pl.cdiv(seq_len, block_q), batch_size, num_heads)

	num_warps_ = num_warps
	if num_warps_ is None:
		num_warps_ = 4 if head_dim <= 64 else 8
	kernel = functools.partial(
		flash_attention_forward_kernel,
		num_heads=num_heads,
		sm_scale=sm_scale,
		block_q=block_q,
		block_k=block_k,
		block_d=head_dim,
	)
	out_shape = [
		jax.ShapeDtypeStruct(shape=query.shape, dtype=query.dtype),  # out
		jax.ShapeDtypeStruct(shape=(batch_size, num_heads, seq_len), dtype=jnp.float32),
		jax.ShapeDtypeStruct(shape=(batch_size, num_heads, seq_len), dtype=jnp.float32),
	]
	in_specs = [
		pl.BlockSpec((None, seq_len, None, head_dim), lambda _, j, k: (j, 0, k, 0)),
		pl.BlockSpec((None, seq_len, None, head_dim), lambda _, j, k: (j, 0, k, 0)),
		pl.BlockSpec((None, seq_len, None, head_dim), lambda _, j, k: (j, 0, k, 0)),
		(
			None
			if bias is None
			else pl.BlockSpec((None, None, seq_len, seq_len), lambda _, j, k: (j, 0, 0, 0))
		),
	]
	out_specs = [
		pl.BlockSpec((None, seq_len, None, head_dim), lambda _, j, k: (j, 0, k, 0)),
		pl.BlockSpec((None, None, seq_len), lambda _, j, k: (j, k, 0)),
		pl.BlockSpec((None, None, seq_len), lambda _, j, k: (j, k, 0)),
	]
	out, l, m = pl.pallas_call(
		kernel,
		grid=grid_,
		in_specs=in_specs,
		out_specs=out_specs,  # type:ignore
		compiler_params=dict(triton=dict(num_warps=num_warps_, num_stages=num_stages)),
		out_shape=out_shape,
		debug=debug,
		interpret=interpret,
		name="flash_attention_forward",
	)(query, key, value, bias)
	return out, (query, key, value, bias, out, l, m)


def _preprocess_backward_kernel(
	out_ref, d_out_ref, l_ref, new_d_out_ref, delta_ref, *, block_q: int
):
	pid_m = pl.program_id(0)

	off_m = pl.ds(pid_m * block_q, block_q)
	# load
	o = pl.load(out_ref, (off_m, slice(None))).astype(jnp.float32)
	do = pl.load(d_out_ref, (off_m, slice(None))).astype(jnp.float32)
	de_num = pl.load(l_ref, (off_m,)).astype(jnp.float32)
	# compute
	do = do / de_num[:, None]
	delta = jnp.sum(o * do, axis=1)
	# write-back
	pl.store(new_d_out_ref, (off_m, slice(None)), do.astype(new_d_out_ref.dtype))
	pl.store(delta_ref, (off_m,), delta.astype(delta_ref.dtype))


@jax.named_scope("preprocess_backward")
def _preprocess_backward(out, do, l, block_q: int, debug: bool, interpret: bool):
	batch_size, seq_len, num_heads, head_dim = out.shape
	out_shape = [
		jax.ShapeDtypeStruct(do.shape, do.dtype),
		jax.ShapeDtypeStruct(l.shape, l.dtype),
	]
	out_specs = [
		pl.BlockSpec((None, seq_len, None, head_dim), lambda _, j, k: (j, 0, k, 0)),
		pl.BlockSpec((None, None, seq_len), lambda _, j, k: (j, k, 0)),
	]
	do_scaled, delta = pl.pallas_call(
		functools.partial(_preprocess_backward_kernel, block_q=block_q),
		grid=(pl.cdiv(seq_len, block_q), batch_size, num_heads),
		in_specs=[
			pl.BlockSpec((None, seq_len, None, head_dim), lambda _, j, k: (j, 0, k, 0)),
			pl.BlockSpec((None, seq_len, None, head_dim), lambda _, j, k: (j, 0, k, 0)),
			pl.BlockSpec((None, None, seq_len), lambda _, j, k: (j, k, 0)),
		],
		out_specs=out_specs,  # type:ignore
		compiler_params=dict(triton=dict(num_warps=4, num_stages=3)),
		out_shape=out_shape,
		debug=debug,
		interpret=interpret,
		name="flash_attention_preprocess_backward",
	)(out, do, l)
	return do_scaled, delta


def flash_attention_backward_kernel(
	# Inputs
	q_ref,
	k_ref,
	v_ref,
	b_ref: jax.Array | None,
	out_ref,
	do_scaled_ref,
	l_ref,
	m_ref,
	delta_ref,
	_,
	# Outputs
	dq_ref,
	dk_ref,
	dv_ref,
	*,
	sm_scale: float,
	block_q: int,
	block_d: int,
	block_k: int,
):
	del out_ref, l_ref  # Not needed
	seq_len = q_ref.shape[0]

	def outer_loop(start_k, _):
		dv = jnp.zeros([block_k, block_d], dtype=jnp.float32)
		dk = jnp.zeros([block_k, block_d], dtype=jnp.float32)
		k = pl.load(k_ref, (pl.ds(start_k * block_k, block_k), slice(None)))
		v = pl.load(v_ref, (pl.ds(start_k * block_k, block_k), slice(None)))

		def inner_loop(start_q, carry):
			dv, dk = carry
			q = pl.load(q_ref, (pl.ds(start_q * block_q, block_q), slice(None)))
			qk = pl.dot(q, k.T)
			qk = qk.astype(q_ref.dtype)
			qk = qk.astype(jnp.float32)
			if sm_scale != 1.0:
				qk *= sm_scale
			if b_ref is not None:
				b = pl.load(
					b_ref,
					(
						pl.ds(start_q * block_q, block_q),
						pl.ds(start_k * block_k, block_k),
					),
				)
				qk = jnp.add(b, qk)

			m = pl.load(m_ref, (pl.ds(start_q * block_q, block_q),))
			p = jnp.exp(qk - m[:, None])
			do = pl.load(do_scaled_ref, (pl.ds(start_q * block_q, block_q), slice(None)))
			dv = dv + pl.dot(p.astype(do.dtype).T, do)
			di = pl.load(delta_ref, (pl.ds(start_q * block_q, block_q),))
			dp = jnp.zeros((block_q, block_k), dtype=jnp.float32) - di[:, None]
			dp = dp + pl.dot(do, v.T)
			ds = p * dp
			if sm_scale != 1.0:
				ds = ds * sm_scale
			dk = dk + pl.dot(ds.astype(q_ref.dtype).T, q)
			dq = pl.load(
				dq_ref,
				(pl.ds(start_q * block_q, block_q), slice(None)),
				eviction_policy="evict_last",
			)
			dq = dq + pl.dot(ds.astype(k.dtype), k).astype(dq.dtype)
			pl.store(
				dq_ref,
				(pl.ds(start_q * block_q, block_q), slice(None)),
				dq,
				eviction_policy="evict_last",
			)
			return dv, dk

		dv, dk = lax.fori_loop(0, pl.cdiv(seq_len, block_q), inner_loop, (dv, dk))
		pl.store(
			dv_ref,
			(pl.ds(start_k * block_k, block_k), slice(None)),
			dv.astype(dv_ref.dtype),
		)
		pl.store(
			dk_ref,
			(pl.ds(start_k * block_k, block_k), slice(None)),
			dk.astype(dk_ref.dtype),
		)

	lax.fori_loop(0, pl.cdiv(seq_len, block_k), outer_loop, None)


@functools.partial(jax.jit, static_argnames=["sm_scale"])
def _flash_attention_reference(
	q,
	k,
	v,
	b: Optional[jax.Array] = None,
	sm_scale=1.0,
):
	logits = jnp.einsum("bqhc,bkhc->bhqk", q, k).astype(jnp.float32)
	logits = b + logits
	weights = jax.nn.softmax(logits * sm_scale).astype(q.dtype)
	return jnp.einsum("bhqk,bkhc->bqhc", weights, v)


def _flash_attention_backward(
	sm_scale: float,
	block_q: int,
	block_k: int,
	backward_pass_impl: str,
	num_warps: int | None,
	num_stages: int,
	grid: Any,
	interpret: bool,
	debug: bool,
	res,
	do,
):
	del num_warps, num_stages, grid
	q, k, v, b, out, l, m = res

	if sm_scale is None:
		sm_scale = 1 / math.sqrt(q.shape[-1])
	if backward_pass_impl == "xla":
		return jax.vjp(
			functools.partial(_flash_attention_reference, sm_scale=sm_scale),
			q,
			k,
			v,
			b,
		)[1](do)
	elif backward_pass_impl == "triton":
		batch_size, seq_len, num_heads, head_dim = q.shape
		block_q = min(block_q, seq_len)
		block_k = min(block_k, seq_len)
		do_scaled, delta = _preprocess_backward(out, do, l, block_q, debug, interpret)
		dq = jnp.zeros(q.shape, jnp.float32)
		out_shapes = [
			jax.ShapeDtypeStruct(dq.shape, dq.dtype),
			jax.ShapeDtypeStruct(k.shape, k.dtype),
			jax.ShapeDtypeStruct(v.shape, v.dtype),
		]

		in_specs = [
			pl.BlockSpec((None, seq_len, None, head_dim), lambda j, k: (j, 0, k, 0)),
			pl.BlockSpec((None, seq_len, None, head_dim), lambda j, k: (j, 0, k, 0)),
			pl.BlockSpec((None, seq_len, None, head_dim), lambda j, k: (j, 0, k, 0)),
			pl.BlockSpec((None, seq_len, None, head_dim), lambda j, k: (j, 0, k, 0)),
			pl.BlockSpec((None, seq_len, None, head_dim), lambda j, k: (j, 0, k, 0)),
			pl.BlockSpec((None, None, seq_len), lambda j, k: (j, k, 0)),
			pl.BlockSpec((None, None, seq_len), lambda j, k: (j, k, 0)),
			pl.BlockSpec((None, None, seq_len), lambda j, k: (j, k, 0)),
			pl.BlockSpec((None, seq_len, None, head_dim), lambda j, k: (j, 0, k, 0)),
		]
		if b is None:
			in_specs.insert(3, None)  # type: ignore[arg-type]
			input_output_aliases = {8: 0}
		else:
			in_specs.insert(
				3,
				pl.BlockSpec(lambda j, k: (j, 0, 0, 0), (None, None, seq_len, seq_len)),
			)
			input_output_aliases = {9: 0}
		grid = (batch_size, num_heads)
		num_warps = 8
		out_specs = [
			pl.BlockSpec((None, seq_len, None, head_dim), lambda j, k: (j, 0, k, 0)),
			pl.BlockSpec((None, seq_len, None, head_dim), lambda j, k: (j, 0, k, 0)),
			pl.BlockSpec((None, seq_len, None, head_dim), lambda j, k: (j, 0, k, 0)),
		]
		dq, dk, dv = pl.pallas_call(
			functools.partial(
				flash_attention_backward_kernel,
				block_q=block_q,
				block_d=head_dim,
				block_k=block_k,
				sm_scale=sm_scale,
			),
			out_specs=out_specs,  # type:ignore
			grid=grid,
			out_shape=out_shapes,
			in_specs=in_specs,
			name="flash_attention_backward",
			debug=debug,
			interpret=interpret,
			compiler_params=dict(triton=dict(num_warps=num_warps, num_stages=1)),
			input_output_aliases=input_output_aliases,
		)(q, k, v, b, out, do_scaled, l, m, delta, dq)
	else:
		raise ValueError(f"Invalid backward pass implementation: {backward_pass_impl}")
	return dq.astype(q.dtype), dk, dv, None


flash_attention.defvjp(_flash_attention_forward, _flash_attention_backward)



"""
__init__ module for custom DataTypes created by fjformer.core
"""

from fjformer.dtypes.arraynf4 import ArrayNF4 as ArrayNF4
from fjformer.dtypes.array8bit import Array8Bit as Array8Bit
from fjformer.dtypes.array8lt import Array8Lt as Array8Lt
